spec CSPwithFC (* This specification is made for solving any kind of
                  Constraint Satisfaction Problem (CSP). 
                  Here, forward-checking is used as solving-strategy
                  and as an example of a CSP the well known eight-queen-
                  problem is given. 
                  The use of cardinalities for productions/transactions
                  and the improvements made by the new "cut-statement"
                  are also shown.   
               *)

   section scheme

      node_class NODE end;

      node_class VARIABLE is_a NODE
         derived
            domain : NODE [0:n];
            actualValue : NODE [0:1];
            isDefined : boolean;
            domainSize : integer = card ( self.domain );
      end;
      (* A Variable is the object which can be bound
         to any value of an given domain
         The specific kind of variable is modelled with
         a node-type of the node-class VARIABLE in the
         problem-specific part of this specification
      *)

      node_class CONSTRAINT is_a NODE end;
      (* A constraint can be declared between two variables.
         The specific kind of constraint is modelled with
         a node-type of the node-class CONSTRAINT in the
         problem-specific part of this specification
      *)

      edge_type ToConstr : VARIABLE [1:1] -> CONSTRAINT [1:n];

      edge_type FromConstr : CONSTRAINT [1:n] -> VARIABLE [1:1];
      (* If there is a constraint 'constr' between two variables
         'var1' and 'var2', which is directed from 'var1' to 'var2'
         then an edge 'ToConstr' points from the corresponding
         node for 'var1' to the node of 'constr'.
         An edge 'FromConstr' is placed between the nodes for
         'constr' and 'var2' further.
      *)

   end;

   section
      MainOperations (* Here, the main operations for solving
                        any kind of CSP by using forward-checking
                        are specified
                     *)
                     (* @CC_LT VertWithIndent @*)

      transaction MAIN [0:1] =
           InitializeEightQueensProblem
         & SolveProblem
         & ShowSolution
      end;
      (* The transaction MAIN must be adapted to the
         specific problem : Initialize the problem and
         solve it.
      *)

      transaction SolveProblem [0:1] =
         use tmpVar : VARIABLE
         do
            loop
               when UndefinedVariableExist
               then
                    SetValueForAnyVariable ( out tmpVar )
                  & SolveConstraintsForVariable ( tmpVar )
            end                                            : [0:1]
         end
      end;
      (* This is the main operation for solving an CSP.
         As long as there is a variable with no defined
         value the following must be done :
         First select some value from the domain and bound
         it to the variable as actual value. After that,
         check if any constraint is hurt.
         The transaction is specified as deterministic,
         because after one solution is found no other one
         is from interest. The transaction is partial,
         because it had to fail if there exists no solution
         for the given problem.
      *)

      transaction SetValueForAnyVariable( out var : VARIABLE) [1:n] =
         use tmpVar : VARIABLE [1:1]
         do
              choose
                 GetOneUndefinedVariable ( out tmpVar )
              else
                 GetAnyUndefinedVariable ( out tmpVar )
              end                                       : [1:1]
            & SetValueForVariable ( tmpVar )
            & var := tmpVar
         end
      end;
      (* This transaction chooses an value from the domain
         of an undefined variable and bounds it to this variable.
         To speed up the runtime-behavior it is first checked
         if there is a variable with only one open value in its
         domain. Only if no such a variable exists any variable
         is selected 'really' nondeterministicly.
         NOTE : The selection is made within a cut-statement
                that declares this part of the transaction as
                total and deterministic !
         The Reasons are : On the one hand it is garanteed that
         there is an undefined variable (when this transaction
         is executed) because of the underlying model. On the
         other hand the solution-finding-process must not select
         a different variable, if the chosen variable can not be
         bounded to a value that holds all constraints. At such a
         point backtracking must be done to the variable that was
         selected right before this variable. For that variable a
         different actual value must be chosen.
         Therefore, the declaration 'deterministic' at the cut-
         statement prevends that 'not needed backtracking'.
      *)

      transaction SetValueForVariable( var : VARIABLE) [1:n] =
         choose
            when ((var.type) = Queen)
            then
               PlaceQueen ( var : Queen )
         else
            halt
         end
         (* This transaction must be adapted to the specific problem.
            Therefore, for each possible type of variable a specific
            transaction for setting the actual value must be defined
         *)
      end;

      transaction SolveConstraintsForVariable( var : VARIABLE) [0:1] =
         use constr : CONSTRAINT [0:n] := nil
         do
              GetConstraintsForVariable ( var, out constr )
            & for_all c : CONSTRAINT := constr
              do
                 SolveConstraint ( c, var )
              end
         end
      end;
      (* Find all constraints for a variable and then try to
         solve them after each other. This transactions is
         partial, because one constraint can be hurt by the
         actual value of the variable and the transaction must
         fail then. It is deterministic, because after solving
         all constraints no other solution for solving exists
         and therefore backtracking here is needlees.
      *)

      transaction SolveConstraint( constr : CONSTRAINT ; var : VARIABLE) [0:1] =
         choose
            when ((constr.type) = NoQueenAttack)
            then
               SolveNoQueenAttack ( constr : NoQueenAttack, var : Queen )
         end
         (* This transaction must be adapted to the specific problem.
            Therefore, for each possible type of constraint a specific
            transaction for solving this constraint must be defined
         *)
      end;

   end;

   section HelpOperations (* In this section operations for working
                             with the underlying model are specified
                          *)

      path NearConstr : VARIABLE -> CONSTRAINT [0:n] =
             ( -ToConstr-> )
         or  ( <-FromConstr- )
      end;
      (* Traversing this path collects all 'constraints'
         that a variable (the starting node) is involved in
      *)

      test UndefinedVariableExist [0:1] =
         begin
            obl_node `1 : VARIABLE;
            (* @CC_LG 6007(60,28)(95,21) @*)
         end
         (* @CC_LG 7008(8,8)(291,47) @*)
         condition `1.isDefined = false;

      end;
      (* Check if an undefined 'variable' exists *)

      test GetOneUndefinedVariable( out v : VARIABLE) [0:n] =
         begin
            obl_node `1 : VARIABLE;
            (* @CC_LG 6007(60,28)(95,21) @*)
         end
         (* @CC_LG 7008(8,8)(291,47) @*)
         condition (`1.isDefined = false) and
                                          (* @CC_LT VertWithSpace @*) (card ( `1.domain ) =1 (* @CC_LT Horizontal @*))
                                                                      (* @CC_LT HoriWithSpace @*)
                                                                      (* @CC_LT HoriWithSpace @*)
                                                                      (* @CC_LT CondVertical @*)
                                                                      (* @CC_LT HoriWithIndent @*)                    ;
         return v := `1;

      end;
      (* Finds an undefined 'variable' that has only
         one possible value in its domain.
         Such 'variables' are preferred within the search
         for an undefined variable
      *)

      test GetAnyUndefinedVariable( out v : VARIABLE) [0:n] =
         begin
            obl_node `1 : VARIABLE;
            (* @CC_LG 6007(60,28)(95,21) @*)
         end
         (* @CC_LG 7008(8,8)(291,47) @*)
         condition (`1.isDefined = false);
         return v := `1;

      end;
      (* Finds any undefined variabale    *)

      test GetConstraintsForVariable( v : VARIABLE ; out c : CONSTRAINT [0:n]) [0:1]
      =
         begin
            obl_node `1 = v;
            (* @CC_LG 6007(36,26)(53,21) @*)
            opt_set `2 : CONSTRAINT;
            (* @CC_LG 6009(216,28)(107,21) @*)
            `1 => `2 : NearConstr;
            (* @CC_LG 3006C @*)
         end
         (* @CC_LG 7008(8,8)(291,47) @*)
         return c := `2;

      end;
      (* Gets all constraints that the given variable
         is involved in
      *)

      production DeleteConstraint( c : CONSTRAINT) [1:1] =
         begin
            obl_node `1 = c;
            (* @CC_LG 6007(48,28)(53,21) @*)
         end
         (* @CC_LG 7008(8,8)(99,47) @*)
         ::=
         (* @CC_LG 7009(8,8)(27,33) @*)
      end;

   end;

   section ProblemSpecificOperations (* Here the problem specfic operation
                                        for the eight-queen-problem are specified
                                     *)

      section EightQueensProblem (* The problem should be clear :
                                    There are eight queens that must be placed on a
                                    chessboard without attacking eachother.
                                    The modelling is as follows :
                                    Every queen is a 'variable' in the CSP and every
                                    square of the chessboard can be one possible position
                                    for the queens. So all squares are 'domain' for the
                                    queens.
                                 *)

         node_type Queen : VARIABLE
            redef_derived
               domain = self.-open->;
               actualValue = self.-at->;
               isDefined = self is with ( -at-> );
         end;
         (* At the beginning all squares of the chessboard
            are in the domain of a queen. By placing one
            queen on the chessboard ('actualValue' then points
            to that square) the domain of all other queens
            must be restrained. This will be done by deleting
            some of the 'open' edges of that queens
         *)

         node_type Square : NODE
            intrinsic
               column : integer;
               row : integer;
         end;
         (* In order to simplify the procedure for testing
            the attack of two queens every square has its
            position within the chessboard as attributes
         *)

         node_type NoQueenAttack : CONSTRAINT end;
         (* This node-type describes the necessity that the
            involved two queens may not attack each other.
            These 'queens' are the nodes at the end/beginning
            of incoming 'ToConstr'/'FromConstr'-edges 
         *)

         edge_type open : VARIABLE [1:n] -> Square [1:n];
         (* Points to all squares that are possible places 
            for the given figure (here : queen)
         *)

         edge_type at : VARIABLE [0:1] -> Square [0:1];
         (* Points to the actual position of the given
            figure on the chessboard
         *)

         edge_type right : Square [0:1] -> Square [0:1];

         edge_type down : Square [0:1] -> Square [0:1];

         edge_type diagUp : Square [0:1] -> Square [0:1];

         edge_type diagDown : Square [0:1] -> Square [0:1];

         path HorizontalAttackedByQueen : Queen [1:1] -> Square [0:n] =
               -at->
            &  (     -right-> +
                 or  <-right- + )
         end;

         path VerticalAttackedByQueen : Queen [1:1] -> Square [0:n] =
               -at->
            &  (     -down-> +
                 or  <-down- + )
         end;

         path DiagonalAttackedByQueen : Queen [1:1] -> Square [0:n] =
               -at->
            &  (     (     -diagDown-> +
                       or  <-diagDown- + )
                 or  (     -diagUp-> +
                       or  <-diagUp- + )   )
         end;

         path ActualPositionFromQueen : Queen [1:1] -> Square [0:1] =
            -at->
         end;

         path AttackedByQueen : Queen [1:1] -> Square [0:n] =
                ActualPositionFromQueen
            or  HorizontalAttackedByQueen
            or  VerticalAttackedByQueen
            or  DiagonalAttackedByQueen
         end;
         (* This path has as result all squares of the 
            chessbaord that are attacked by the given
            queen. The square that is the actual position
            of the queen is included, because no other
            queen can be set on this position and therefore
            it must be cleared out of the domain of the
            other queens, too. 
         *)

         function .. : ( a, b : integer) -> integer [0:n] =
            [ a >= b :: a
            | a or ((a + 1) .. b) ]
         end;

         production ShowSolution [1:1] =
            begin
               opt_set `1 : Square;
               (* @CC_LG 6009(145,72)(97,29) @*)
               => `1 : not with ( <-at- )
                                (* @CC_LG 4005LC,RC(-8,0)(108,12) @*);
               (* @CC_LG
               3009XC2
               144,28 @*)
            end
            (* @CC_LG 7008(8,8)(291,103) @*)
            ::=
            (* @CC_LG 7009(8,8)(291,47) @*)
         end;
         (* This productions removes all 'square'-nodes that
            are not be placed with a queen. The result is that
            after finding a solution for a general n-queen-
            problem the correctness of this solution can be
            verified easier
         *)

         section SolvingOperations

            production PlaceQueen( theQueen : Queen) [1:n] =
               begin
                  obl_node `1 = theQueen;
                  (* @CC_LG 6007(60,28)(95,21) @*)
                  obl_node `3 : Square;
                  (* @CC_LG 6007(60,112)(83,21) @*)
                  `1 -> `3 : open;
                  (* @CC_LG 3004 @*)
                  opt_node `2 : Square;
                  (* @CC_LG 6008(228,28)(83,21) @*)
                  `1 -> `2 : at;
                  (* @CC_LG 3004C @*)
               end
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(60,28)(59,21) @*)
                  obl_node 3' = `3;
                  (* @CC_LG 6007(60,112)(59,21) @*)
                  1' -> 3' : at;
                  (* @CC_LG 3004C @*)
                  opt_node 2' = `2;
                  (* @CC_LG 6008(220,28)(59,21) @*)
               end
               (* @CC_LG 7009(8,8)(291,131) @*)
            end;
            (* This production places a queen on one of the
               possible places on the chessboard.
               It is total, because of the forward-checking strategy
               it is guaranteed that every queen has as least one
               possible open place in its domain.
               It must be nondeterministic, because a different place
               must be chosen via backtracking if one constraint is hurt
            *)

            production DeleteOpenPositionsForQueen( theQueen : Queen ; thePositions : Square [0:n])
               * =
               begin
                  obl_node `1 = theQueen;
                  (* @CC_LG 6007(60,42)(95,21) @*)
                  obl_node `2 = elem ( thePositions );
                  (* @CC_LG 6007(232,42)(173,21) @*)
                  `1 -> `2 : open;
                  (* @CC_LG 3004C @*)
               end
               (* @CC_LG 7008(8,8)(327,75) @*)
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(48,23)(59,21) @*)
                  obl_node 2' = `2;
                  (* @CC_LG 6007(196,23)(59,21) @*)
               end
               (* @CC_LG 7009(8,8)(303,47) @*)
            end;
            (* If one queen is placed on the chessboard this production
               deletes the attacked squares out of the domain of the given
               queen, because they are no longer possible positions for 
               that queen 
            *)

            test GetPositionsAttackedByQueen( theQueen : Queen ; out positions : Square [0:n])
               [1:1] =
               begin
                  obl_node `1 = theQueen;
                  (* @CC_LG 6007(60,28)(95,21) @*)
                  obl_set `2 : Square;
                  (* @CC_LG 6010(264,28)(83,21) @*)
                  `1 => `2 : AttackedByQueen;
                  (* @CC_LG 3006C @*)
               end
               (* @CC_LG 7008(8,8)(363,47) @*)
               return positions := `2;

            end;
            (* Find all squares that are attacked by the given queen *)

            transaction SolveNoQueenAttack( constr : NoQueenAttack ; sourceQueen : Queen) [0:1]
            =
               use targetQueen : Queen;
                   targetDomain : NODE [0:n];
                   targetSize : integer;
                   attackedPositions : Square [0:n] := nil
               do
                    choose
                       when (constr in (sourceQueen.-ToConstr->))
                       then
                          targetQueen := (constr.-FromConstr->) : Queen
                    else
                       targetQueen := (constr.<-ToConstr-) : Queen
                    end                                                 : [1:1]
                  & begin
                         GetPositionsAttackedByQueen ( sourceQueen, out attackedPositions )
                       & DeleteOpenPositionsForQueen ( targetQueen, attackedPositions )
                    end                                                                     : [1:1]
                  & targetDomain := targetQueen.domain
                  & targetSize := targetQueen.domainSize
                  & choose
                       when (targetQueen.domainSize = 0)
                       then
                          fail
                    else
                       DeleteConstraint ( constr )
                    end
               end
            end;
            (* This is the main transaction that solves a constraint 
               between two queens. As parameters there are the (source) 
               queen that was assigned an actual value (position on the
               board) and the constraint to one other queen.
               Because the constraints are directed (which is not nessessary
               for this problem, but holds for general problems) the involved
               'other queen' (target queen) must be determined first. Then,
               the attacked positions must be find out and be deleted out of
               the domain of the target queen. After that, it is checked if
               the domainsize is zero. If so, the target queen can't be set
               to the chessboard without hurting this constraint. In this case,
               the given source queen must be set to a different positions.
               At this point the transaction fails and because the results
               determined by the previous commands are really deterministic,
               no backtracking is needed there. 
               On the other hand if the given constraint can be hold it is
               deleted. This prevends the twice selecting of the constraint.
            *)

         end;

         section ChessboardBuildingOperations (* The following operations build up the chessboard
                                                 for the eight-queen-problem.
                                                 There is a great use of '*/+'-productions !
                                              *)

            production CreateSingleSquare [1:1] =
               (* @CC_LG 7008(8,8)(24,24) @*)
               ::=
               begin
                  1' : Square;
                  (* @CC_LG 6011(48,28)(89,21) @*)
               end
               (* @CC_LG 7009(8,8)(99,47) @*)
               transfer 1'.column := 1;
                        1'.row := 1;
            end;

            production ExtendColumn * =
               begin
                  obl_node `1 : Square;
                  (* @CC_LG 6007(60,27)(83,21) @*)
                  not_node `2 : Square;
                  (* @CC_LG 6012(60,90)(83,21) @*)
                  `1 -> `2 : down;
                  (* @CC_LG 3004C @*)
               end
               (* @CC_LG 7008(8,8)(135,117) @*)
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(60,26)(59,21) @*)
                  2' : Square;
                  (* @CC_LG 6011(60,84)(83,21) @*)
                  1' -> 2' : down;
                  (* @CC_LG 3004C @*)
               end
               (* @CC_LG 7009(8,8)(135,131) @*)
               transfer 2'.column := `1.column;
                        2'.row := (`1.row + 1);
            end;

            production CloseColumn * =
               begin
                  obl_node `1 : Square;
                  (* @CC_LG 6007(60,28)(83,21) @*)
                  obl_node `2 : Square;
                  (* @CC_LG 6007(228,28)(83,21) @*)
                  obl_node `3 : Square;
                  (* @CC_LG 6007(60,112)(83,21) @*)
                  obl_node `4 : Square;
                  (* @CC_LG 6007(228,112)(83,21) @*)
                  `1 -> `3 : down;
                  (* @CC_LG 3004 @*)
                  `1 -> `2 : right;
                  (* @CC_LG 3004 @*)
                  `3 -> `4 : right;
                  (* @CC_LG 3004 @*)
                  `2 +> `4 : down;
                  (* @CC_LG 3008 @*)
               end
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(60,28)(59,21) @*)
                  obl_node 2' = `2;
                  (* @CC_LG 6007(228,28)(59,21) @*)
                  obl_node 3' = `3;
                  (* @CC_LG 6007(60,112)(59,21) @*)
                  obl_node 4' = `4;
                  (* @CC_LG 6007(228,112)(59,21) @*)
                  1' -> 3' : down (* @CC_LG 1002C,BC(0,-10)(24,10) @*);
                  (* @CC_LG 3004C @*)
                  1' -> 2' : right (* @CC_LG 1002C,BC(0,-10)(30,10) @*);
                  (* @CC_LG 3004C @*)
                  3' -> 4' : right (* @CC_LG 1002C,BC(0,-10)(30,10) @*);
                  (* @CC_LG 3004C @*)
                  2' -> 4' : down;
                  (* @CC_LG 3004C @*)
               end
            end;

            production ExtendRow * =
               begin
                  obl_node `1 : Square;
                  (* @CC_LG 6007(60,28)(83,21) @*)
                  not_node `2 : Square;
                  (* @CC_LG 6012(228,28)(83,21) @*)
                  `1 -> `2 : right;
                  (* @CC_LG 3004 @*)
               end
               (* @CC_LG 7008(8,8)(291,47) @*)
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(60,28)(59,21) @*)
                  2' : Square;
                  (* @CC_LG 6011(228,28)(83,21) @*)
                  1' -> 2' : right;
                  (* @CC_LG 3004 @*)
               end
               (* @CC_LG 7009(8,8)(291,47) @*)
               transfer 2'.column := (`1.column + 1);
                        2'.row := `1.row;
            end;

            production CloseRow * =
               begin
                  obl_node `1 : Square;
                  (* @CC_LG 6007(60,28)(83,21) @*)
                  obl_node `2 : Square;
                  (* @CC_LG 6007(228,28)(83,21) @*)
                  obl_node `3 : Square;
                  (* @CC_LG 6007(60,112)(83,21) @*)
                  obl_node `4 : Square;
                  (* @CC_LG 6007(228,112)(83,21) @*)
                  `1 -> `3 : down;
                  (* @CC_LG 3004 @*)
                  `1 -> `2 : right;
                  (* @CC_LG 3004 @*)
                  `2 -> `4 : down;
                  (* @CC_LG 3004 @*)
                  `3 +> `4 : right;
                  (* @CC_LG 3008 @*)
               end
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(60,28)(59,21) @*)
                  obl_node 2' = `2;
                  (* @CC_LG 6007(228,28)(59,21) @*)
                  obl_node 3' = `3;
                  (* @CC_LG 6007(60,112)(59,21) @*)
                  obl_node 4' = `4;
                  (* @CC_LG 6007(228,112)(59,21) @*)
                  1' -> 3' : down (* @CC_LG 1002C,BC(0,-10)(24,10) @*);
                  (* @CC_LG 3004C @*)
                  1' -> 2' : right (* @CC_LG 1002C,BC(0,-10)(30,10) @*);
                  (* @CC_LG 3004C @*)
                  2' -> 4' : down (* @CC_LG 1002C,BC(0,-10)(24,10) @*);
                  (* @CC_LG 3004C @*)
                  3' -> 4' : right;
                  (* @CC_LG 3004C @*)
               end
            end;

            production CloseDiag * =
               begin
                  obl_node `1 : Square;
                  (* @CC_LG 6007(60,28)(83,21) @*)
                  obl_node `2 : Square;
                  (* @CC_LG 6007(228,28)(83,21) @*)
                  obl_node `3 : Square;
                  (* @CC_LG 6007(60,112)(83,21) @*)
                  obl_node `4 : Square;
                  (* @CC_LG 6007(228,112)(83,21) @*)
                  `1 -> `3 : down;
                  (* @CC_LG 3004 @*)
                  `1 -> `2 : right;
                  (* @CC_LG 3004 @*)
                  `2 -> `4 : down (* @CC_LG 1002C,BC(0,32)(24,10) @*);
                  (* @CC_LG 3004 @*)
                  `3 -> `4 : right;
                  (* @CC_LG 3004 @*)
                  `3 +> `2 : diagUp (* @CC_LG 1002C,BC(47,7)(36,10) @*);
                  (* @CC_LG 3008 @*)
                  `1 +> `4 : diagDown (* @CC_LG 1002C,BC(-47,2)(48,10) @*);
                  (* @CC_LG 3008 @*)
               end
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(60,28)(59,21) @*)
                  obl_node 2' = `2;
                  (* @CC_LG 6007(228,28)(59,21) @*)
                  obl_node 3' = `3;
                  (* @CC_LG 6007(60,112)(59,21) @*)
                  obl_node 4' = `4;
                  (* @CC_LG 6007(228,112)(59,21) @*)
                  1' -> 3' : down (* @CC_LG 1002C,BC(0,-10)(24,10) @*);
                  (* @CC_LG 3004C @*)
                  1' -> 2' : right (* @CC_LG 1002C,BC(0,-10)(30,10) @*);
                  (* @CC_LG 3004C @*)
                  2' -> 4' : down (* @CC_LG 1002C,BC(0,32)(24,10) @*);
                  (* @CC_LG 3004C @*)
                  3' -> 4' : right;
                  (* @CC_LG 3004C @*)
                  3' -> 2' : diagUp (* @CC_LG 1002C,BC(42,6)(36,10) @*);
                  (* @CC_LG 3004 @*)
                  1' -> 4' : diagDown (* @CC_LG 1002C,BC(-48,2)(48,10) @*);
                  (* @CC_LG 3004 @*)
               end
            end;
            (* This productions is necessary for creating diagonal
               edges inside the 'net' of squares. They are needed
               to simplify the diagonal moving of the queens.      
            *)

            transaction CreateChessboard( n : integer) [1:1] =
                 CreateSingleSquare
               & for_all i : integer := (2 .. n)
                 do
                      ExtendRow
                    & CloseColumn
                    & ExtendColumn
                    & CloseRow
                 end
               & CloseDiag
            end;

         end;

         section QueenCreationOperations

            production CreateOneQueen( out newQueen : Queen) [1:1] =
               (* @CC_LG 7008(8,8)(24,24) @*)
               ::=
               begin
                  1' : Queen;
                  (* @CC_LG 6011(48,28)(89,21) @*)
               end
               (* @CC_LG 7009(8,8)(99,47) @*)
               return newQueen := 1';
            end;

            production SetDomainForQueen( theQueen : Queen) [1:1] =
               begin
                  obl_node `1 = theQueen;
                  (* @CC_LG 6007(59,28)(95,21) @*)
                  obl_set `2 : Square;
                  (* @CC_LG 6010(228,28)(83,21) @*)
               end
               (* @CC_LG 7008(8,8)(339,47) @*)
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(46,28)(59,21) @*)
                  obl_set 2' = `2;
                  (* @CC_LG 6010(220,28)(59,21) @*)
                  1' -> 2' : open;
                  (* @CC_LG 3004C @*)
               end
               (* @CC_LG 7009(8,8)(339,47) @*)
            end;
            (*  In the beginning all squares of the chessboard
                are possible positions for every queen
            *)

            production CreateConstraintsForQueens * =
               begin
                  obl_node `1 : Queen;
                  (* @CC_LG 6007(60,95)(95,21) @*)
                  obl_node `2 : Queen;
                  (* @CC_LG 6007(324,95)(83,21) @*)
                  not_node `3 : NoQueenAttack;
                  (* @CC_LG 6012(192,168)(131,21) @*)
                  `2 -> `3 : ToConstr (* @CC_LG 1002C,BC(48,2)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  `3 -> `1 : FromConstr (* @CC_LG 1002C,BC(-42,2)(60,10) @*);
                  (* @CC_LG 3004C @*)
                  not_node `4 : NoQueenAttack;
                  (* @CC_LG 6012(192,28)(131,21) @*)
                  `4 -> `2 : FromConstr (* @CC_LG 1002C,BC(48,-7)(60,10) @*);
                  (* @CC_LG 3004C @*)
                  `1 -> `4 : ToConstr (* @CC_LG 1002C,BC(-42,-10)(48,10) @*);
                  (* @CC_LG 3004 @*)
               end
               (* @CC_LG 7008(8,8)(375,187) @*)
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(36,42)(51,21) @*)
                  obl_node 2' = `2;
                  (* @CC_LG 6007(336,42)(59,21) @*)
                  4' : NoQueenAttack;
                  (* @CC_LG 6011(180,44)(131,21) @*)
                  4' -> 2' : FromConstr (* @CC_LG 1002(-5,-29)(60,10) @*);
                  (* @CC_LG 3004C @*)
                  1' -> 4' : ToConstr (* @CC_LG 1002(-1,-28)(48,10) @*);
                  (* @CC_LG 3004C @*)
               end
               (* @CC_LG 7009(8,8)(375,68) @*)
            end;
            (* This productions creates one constraint between two
               queens if such one does not still exist.
               Note : For a correct modelling of the eight-queen-problem
               there must be two inverse directed constraints (Every two
               queen can attack each other). But with solving the first
               constraint the second one is solved, too. Because of that
               the second constraint is not modelled here and the solving
               transaction 'SolveNoQueenAttack' considers that.
            *)

            transaction CreateNQueens( n : integer) [1:1] =
               use num : integer := n;
                   newQueen : Queen
               do
                  loop
                     when (num > 0)
                     then
                          CreateOneQueen ( out newQueen )
                        & SetDomainForQueen ( newQueen )
                        & num := (num - 1)
                  end
               end
            end;
            (* Creates 'queen'-nodes for the given number
               of queens and initializes the domains
            *)

         end;

      end;

      transaction InitializeEightQueensProblem [1:1] =
           CreateChessboard ( 8 )
         & CreateNQueens ( 8 )
         & CreateConstraintsForQueens
      end;

   end;

end.
