spec FerryMan (* This programmed graph rewriting system is a rule based   *)
              (* specification of the following well-known AI problem:    *)
              (*                                                          *)
              (* The ferry man's problem:                                 *)
              (* ------------------------                                 *)
              (*     There are a cabbage, a goat, and a wolf which must   *)
              (*     be ferried over a small river. But the ferry is too  *)
              (*     small to carry more than one of these objects at a   *)
              (*     time. Furthermore, the ferry man must not leave the  *)
              (*     wolf and the goat or the goat and the cabbage at one *)
              (*     river side with himself being on the other side.     *)

   section AuxiliaryFunctions

      function $ : ( Dollar : integer) -> integer =
         Dollar * 100
      end;

      function Payment : -> integer =
         $ 30
      end;

   end;
   (* Auxiliary Functions *)

   section MainActions

      transaction SolveFerryMansProblem (* Try&Error solution for the ferry  *)
                                        (* man's problem. We are looking for *)
                                        (* a solution with a minimal number  *)
                                        (* of basic operations.              *) =
         use StepCounter : integer
         do
              CreateProblem
            & Counter := 0
            & loop
                 when def ( StillCargoOnThisSide )
                 then
                      OneStep
                    & Counter := (Counter + 1)
              end
         end
      end;

      transaction OneStep =
            TransferCargo
         or EmptyMove
      end;

      transaction TransferCargo =
           LoadFerry
         & not DangerousSituation
         & FerryOver
         & UnloadFerry
      end;

      transaction EmptyMove =
           not DangerousSituation
         & FerryOver
      end;

   end;
   (* SolveFerryMansProblem *)

   section GraphScheme (* Definition of the overall graph scheme *)
                       (* for boats, cargos etc.                 *)

      node_class THING
         intrinsic
            Expenditures := $ 0;
         derived
            Price = $ 0;
            Win = self.Price - self.Expenditures;
      end;

      node_class RIVER_SIDE end;

      node_class CARGO is_a THING end;

      edge_type On : CARGO [0:n] -> RIVER_SIDE [0:1];
      (* A CARGO is at most at one river side and we have *)
      (* an arbitrary number of cargos on each river side *)

      edge_type In : CARGO -> BOAT [0:1];
      (* A CARGO is in at most one BOAT at the same time *)

      node_class BOAT is_a THING
         intrinsic
            Income := $ 70;
         derived
            TransferCosts = ($ 20) + insurance ( [ self.<-In-.Price ] );
            (* TransferCosts are fixed costs plus insurance which is *)
            (* proportional to the price of the BOAT'S CARGO         *)
         redef_derived
            Win = self.Income - self.Expenditures;
      end;

      edge_type At : BOAT [0:n] -> RIVER_SIDE [1:1];

      node_class BEING is_a THING end;

      node_class PLANT is_a BEING end;

      node_class ANIMAL is_a BEING end;

   end;
   (* GlobalScheme *)

   section SpecialObjectTypes (* Definition of the problem specific *)
                              (* object types Ferry, Wolf etc.      *)

      node_type ThisSide : RIVER_SIDE end;

      node_type OtherSide : RIVER_SIDE end;

      node_type Ferry : BOAT
         redef_derived
            Price = $ 1000;
      end;

      node_type Cabbage : PLANT
         redef_derived
            Price = $ 40;
      end;

      node_type Goat : ANIMAL
         redef_derived
            Price = $ 100;
      end;

      node_type Wolf : ANIMAL
         redef_derived
            Price = $ 60;
      end;

   end;
   (* SpecialObjectTypes *)

   section BasicOperations (* Definition of all legal basic actions of *)
                           (* the ferry man - like 'LoadFerry' - and   *)
                           (* some auxiliary graph rewrite rules.      *)

      production CreateProblem =
         (<OptPrecondDecl>)
         (* Creates the initial situation with all *)
         (* objects on this side of the river.     *)
         (* @CC_LG 7008(8,8)(26,28) @*)
         (<OptLeftSideList>)
         (* @CC_LG 7008C(27,33) @*)
         ::=
         begin
            1' : Cabbage;
            (* @CC_LG 6011(228,28)(89,21) @*)
            2' : Wolf;
            (* @CC_LG 6011(48,28)(71,21) @*)
            3' : Goat;
            (* @CC_LG 6011(132,28)(71,21) @*)
            4' : ThisSide;
            (* @CC_LG 6011(132,146)(95,21) @*)
            5' : Ferry;
            (* @CC_LG 6011(250,146)(77,21) @*)
            6' : OtherSide;
            (* @CC_LG 6011(364,146)(101,21) @*)
            2' -> 4' : On;
            (* @CC_LG 3004C @*)
            3' -> 4' : On;
            (* @CC_LG 3004C @*)
            1' -> 4' : On;
            (* @CC_LG 3004C @*)
            5' -> 4' : At;
            (* @CC_LG 3004C @*)
         end
         (* @CC_LG 7009(8,8)(427,177) @*)
      end;

      production LoadFerry (* Puts one cargo into the ferry    *)
                           (* if the ferry is on the same side *)
                           (* and empty.                       *) =
         begin
            obl_node `1 : CARGO;
            (* @CC_LG 6007(65,26)(77,21) @*)
            obl_node `2 : RIVER_SIDE;
            (* @CC_LG 6007(65,126)(107,21) @*)
            obl_node `3 : Ferry;
            (* @CC_LG 6007(250,126)(77,21) @*)
            `1 -> `2 : On;
            (* @CC_LG 3004C @*)
            `3 -> `2 : At;
            (* @CC_LG 3004C @*)
            => `3 : WithoutCargo;
            (* @CC_LG
            3009
            252,84 @*)
         end
         (* @CC_LG 7008(8,8)(305,157) @*)
         ::=
         begin
            obl_node 1' = `1;
            (* @CC_LG 6007(41,40)(59,21) @*)
            obl_node 2' = `2;
            (* @CC_LG 6007(41,140)(59,21) @*)
            obl_node 3' = `3;
            (* @CC_LG 6007(226,140)(59,21) @*)
            3' -> 2' : At (* @CC_LG 1002C,BC(0,-9)(12,10) @*);
            (* @CC_LG 3004C @*)
            1' -> 3' : In;
            (* @CC_LG 
            3004C3
            226,40 @*)
         end
         (* @CC_LG 7009(8,8)(269,171) @*)
         condition `1.Win > Payment;
         transfer 1'.Expenditures := `1.Expenditures + Payment;
                  3'.Income := `3.Income + Payment;
      end;

      production FerryOver =
         begin
            obl_node `1 : RIVER_SIDE;
            (* @CC_LG 6007(71,40)(107,21) @*)
            obl_node `2 : Ferry;
            (* @CC_LG 6007(204,40)(77,21) @*)
            obl_node `3 : RIVER_SIDE;
            (* @CC_LG 6007(333,40)(107,21) @*)
            `2 -> `1 : At;
            (* @CC_LG 3004C @*)
         end
         (* @CC_LG 7008(8,8)(401,71) @*)
         ::=
         begin
            obl_node 1' = `1;
            (* @CC_LG 6007(47,40)(59,21) @*)
            obl_node 2' = `2;
            (* @CC_LG 6007(180,40)(59,21) @*)
            obl_node 3' = `3;
            (* @CC_LG 6007(309,40)(59,21) @*)
            2' -> 3' : At;
            (* @CC_LG 3004C @*)
         end
         (* @CC_LG 7009(8,8)(353,71) @*)
         condition `2.Win > `2.TransferCosts;
         transfer 2'.Expenditures := `2.Expenditures + `2.TransferCosts;
      end;

      test StillCargoOnThisSide =
         begin
            obl_node `1 : CARGO;
            (* @CC_LG 6007(64,26)(77,21) @*)
            obl_node `2 : ThisSide;
            (* @CC_LG 6007(64,132)(95,21) @*)
            `1 -> `2 : On;
            (* @CC_LG 3004C @*)
         end
         (* @CC_LG 7008(8,8)(125,157) @*)

      end;

      test DangerousSituation =
         begin
            obl_node `2 : BEING;
            (* @CC_LG 6007(260,40)(77,21) @*)
            obl_node `1 : ANIMAL;
            (* @CC_LG 6007(56,41)(83,21) @*)
            `1 => `2 : NearByFood;
            (* @CC_LG 3006C @*)
         end
         (* @CC_LG 7008(8,8)(315,71) @*)

      end;

   end;
   (* BasicOperations *)

   section DerivedRelations (* Definition of complex relations and *)
                            (* restrictions.                       *)

      path NearByFood : ANIMAL -> BEING =
         [ NearBy ( Wolf, Goat )
         | NearBy ( Goat, Cabbage ) ]
      end;

      path NearBy( Subj : type_in ANIMAL ; Obj : type_in BEING) : ANIMAL -> BEING =
            instance_of Subj
         &  -On->
         &  <-On-
         &  instance_of Obj
      end;

      restriction WithoutCargo : BOAT =
         not with <-In-
      end;

   end;
   (* DerivedRelations *)

   transaction MAIN =
      SolveFerryMansProblem
   end;

end.
