/***************/
/* FileUtils.c */
/***************/

/**************************************************/
/* implementation of imported types and functions */
/**************************************************/

/***************************************************************************/
/*                                                                         */
/*  ATTENTION: please SET environment Variable "IMPORTLIB" according       */
/*             to the library containing the object code of the            */
/*             imported types and functions.                               */
/*                                                                         */
/***************************************************************************/

/* Note further that an imported type needs to be provided with at 
   least three pre-defined functions that are needed for PROGRES built-in 
   functionality.
   * (let USERTYPE be the type to be imported)
   *  - USERTYPE_Size returns the size of the type in the parameter
   *  - USERTYPE_ValueToString writes the contents of the imported type to a string, 
   *    that will be returned in the first parameter
   *  - USERTYPE_StringToValue converts a string to USERTYPE and returns it by
   *    parameter
   * Remark: The formerly needed procedures WriteConst and ReadConst are obsolete!

   * Further user-defined functions for an imported type have the following rules: 
   * Assume that you want to import a function in the specification:
   *   MyImportedFunction : (T1, ..., Tn) -> Tr
   * where types T1, ..., Tn, Tr are imported or built-in types.
   * The procedure implementing the function above must then look like:
   *   PROCEDURE MyImportedFunction (Tr *pr, T1 *p1, ..., Tn *pn);
   * p1, ..., pn are the in-paramters and pr is the result of the function.
   *
   * Note: All parameters must be passed by address (for technical reasons)
 */


#include <stdio.h>
#include <stdlib.h>

/* use type definition of text in the header file */
#include "FileUtils.h"


/* header of local functions */


/***********************************************************/
/* implementation of standard functions for imported types */
/***********************************************************/

/* The following three procedures are obligatory in connection 
 * with an imported type. They must not be used in the specification
 * explicitely, but they are needed for internal use (e.g. conversion
 * from text to string and vice versa. */
/* Note: The procedure names and the parameters (type and order) are important.
 * The convention is, that in procedures results are returned as the first  
 * parameter. */

void FUFile_Size (int *size)
/* tell the size of file type */
{
	*size = sizeof (FUFile);
}


void FUFile_ValueToString (PGCMString *outString, FUFile *f)
/* string representation for type FUFile */
{
	strcpy ((*outString).s, "file type");
}


void FUFile_StringToValue (FUFile *f, PGCMString *inString)
/* Make file out of string - use FUOpenFile instead */
{
	f = NULL;
}


/*****************************************************/
/* implementation of user-defined imported functions */
/*****************************************************/

void FUOpenFile (FUFile *f, PGCMString *fileName)
/* Open a file of given name. If fopen fails NULL is returned. */
{
  FILE *hFile;

  hFile = fopen ((*fileName).s, "r");
  if (!hFile) {
    fprintf (stderr, "FileUtils: fopen failed for %s.\n", (*fileName).s);
  }
  *f = hFile;
}


void FUCreateFile (FUFile *f, PGCMString *fileName)
/* Open a file of given name. If fopen fails NULL is returned. */
{
  FILE *hFile;

  hFile = fopen ((*fileName).s, "w");
  if (!hFile) {
    fprintf (stderr, "FileUtils: fopen failed for %s.\n", (*fileName).s);
  }
  *f = hFile;
}

void FUIsFileOpen (BOOLEAN *s, FUFile *f)
/* Return BOOLEAN corresponding to f being a valid file pointer. */
/* Error message to sderr may also be suppressed. */
{
  FILE *hFile;

/*
  if (*f) {
    fprintf (stderr, "FileUtils: File is open.\n");
  } else {
    fprintf (stderr, "FileUtils: File is not open.\n");
  }
*/

  *s = (*f != NULL);
}


void FUCloseFile (BOOLEAN *s, FUFile *f)
/* Closing files after use is okay. */
{
  fclose (*f);
  *s = TRUE;
}


void FUIsEOF (BOOLEAN *s, FUFile *f)
/* Is end-of-file flag set for file? */
{
  *s = feof (*f);
}


void FUReadLine (PGCMString *outString, FUFile *f)
/* Read until end-of-line (if shorter than MAXLINELENGTH). */
{
  char fileName[100];
  FILE *hFile;
  char buffer[MAXLINELENGTH];
  char *newline;
  char *res;
  int i;

  res = fgets (buffer, MAXLINELENGTH, *f);

  if (!res) {
    (*outString).s[0] = '\0';
#ifdef DEBUG
    fprintf (stderr, "FileUtils: Couldn't read.\n");
#endif
    return;
  }
  /*  printf("Found: %s\n", buffer); */

  if ((newline = strchr (buffer, '\n')) != NULL) {
    *newline = '\0';
  }

  i = 0;
  while (i <= MAXLINELENGTH) {
    (*outString).s[i] = buffer[i];
    if (i == PGCMStringLength - 1) {
      (*outString).s[i] = '\0';
      i = MAXLINELENGTH;
    }
    i++;
  }
}

void FUWriteLine (BOOLEAN *b, FUFile *f, PGCMString *str) {
  char *res;

  res = fprintf (*f, "%s\n", (*str).s);
  if (!res) {
    fprintf (stderr, "FileUtils: Couldn't write line.\n");
    *b = FALSE;
    return;
  }
  *b = TRUE;
}


void FUWriteText (BOOLEAN *b, FUFile *f, PGCMString *str) {
  char *res;

  res = fprintf (*f, "%s", (*str).s);
  if (!res) {
    fprintf (stderr, "FileUtils: Couldn't write text.\n");
    *b = FALSE;
    return;
  }
  *b = TRUE;
}


void FUReadString (PGCMString *outString, FUFile *f)
/* Read string until whitespace is read. */
{
  char fileName[100];
  FILE *hFile;
  char buffer[MAXLINELENGTH];
  char *newline;
  char *res;
  int i;
  char c;
  int leading_ws;
  
  i = 0;
  /* initialize empty string */
  (*outString).s[i] = '\0';

  c = fgetc (*f);
  leading_ws = 1;
  while (!feof (*f) && (i < PGCMStringLength)) {
    if (feof (*f)) {
      /* stop reading */
      (*outString).s[i] = '\0';
      i = PGCMStringLength;
    } else if (((c == ' ') || (c == '\n')) && leading_ws) {
      /* just skip whitespace read before characters */
      c = fgetc (*f);
    } else if ((c == ' ') || (c == '\n')) {
      /* stop reading when whitespace is found */
      (*outString).s[i] = '\0';
      i = PGCMStringLength;
    } else {
      /* characters to read */
      leading_ws = 0;
      (*outString).s[i] = c;
      i++;
      c = fgetc (*f);
    }
  }
  
  (*outString).s[PGCMStringLength - 1] = '\0';
}


void FUReadNumber (int *outNum, FUFile *f)
/* Read until whitespace and convert the found string as numeral value. */
{
  char fileName[100];
  FILE *hFile;
  char buffer[MAXLINELENGTH];
  char *newline;
  char *res;
  int i;
  char c;
  int leading_ws;

  i = 0;
  res = 0;
  leading_ws = 1;
  c = fgetc (*f);
  while (!feof (*f) && (i < PGCMStringLength)) {
    if (feof (*f)) {
      /* stop reading */
      buffer[i] = '\0';
      i = PGCMStringLength;
    } else if (((c == ' ') || (c == '\n')) && leading_ws) {
      /* just skip whitespace read before number */
      c = fgetc (*f);
    } else if (isdigit (c)) {
      /* start reading number */
      leading_ws = 0;
      buffer[i] = c;
      i++;
      c = fgetc (*f);
    } else {
      /* stop reading when whitespace or other character is found */
      buffer[i] = '\0';
      i = PGCMStringLength;
    }
  }
  
  *outNum = atoi (buffer);
}

