/*****************/
/* LongStrings.c */
/*****************/

/**************************************************/
/* implementation of imported types and functions */
/**************************************************/

/***************************************************************************/
/*                                                                         */
/*  ATTENTION: please SET environment Variable "IMPORTLIB" according       */
/*             to the library containing the object code of the            */
/*             imported types and functions.                               */
/*                                                                         */
/***************************************************************************/

/* Note further that an imported type needs to be provided with at 
 * least three pre-defined functions that are needed for PROGRES built-in 
 * functionality.
 * (let USERTYPE be the type to be imported)
 *  - USERTYPE_Size returns the size of the type in the parameter
 *  - USERTYPE_ValueToString writes the contents of the imported type 
 *    to a string, that will be returned in the first parameter
 *  - USERTYPE_StringToValue converts a string to USERTYPE and returns it by
 *    parameter
 * Remark: The formerly needed procedures WriteConst and ReadConst 
 * are obsolete!

 * Further user-defined functions for an imported type have the following 
 * rules: 
 * Assume that you want to import a function in the specification like
 *   MyImportedFunction : (T1, ..., Tn) -> Tr
 * where types T1, ..., Tn, Tr are imported or built-in types.
 * The procedure implementing the function above must then look like:
 *   PROCEDURE MyImportedFunction (Tr *pr, T1 *p1, ..., Tn *pn);
 * p1, ..., pn are the in-paramters and pr is the result of the function.
 *
 * Note: All parameters must be passed by address (for technical reasons)
 */


#include <stdio.h>
#include <stdlib.h>

/* use type definition of text in the header file */
#include "LongStrings.h"


/***********************************************************/
/* implementation of standard functions for imported types */
/***********************************************************/

/* The following three procedures are obligatory in connection 
 * with an imported type. They must not be used in the specification
 * explicitely, but they are needed for internal use (e.g. conversion
 * from text to string and vice versa. 
 */
/* Note: The procedure names and the parameters (type and order) 
 * are important. The convention is, that in procedures results are 
 * returned as the first parameter. 
 */

void text_Size (int *size)
/* tell only the size of the real textbuffer in type text */
{
  *size = sizeof (text);
}


void text_ValueToString (PGCMString *outString, text *value)
/* convert text to string (void) */
{
  int i, len;

  len = (value->curLength < PGCMStringLength) ? value->curLength
    : PGCMStringLength;
  for (i = 0; i < len; i++)
    {
      (*outString).s[i] = value->contents[i];
    }
  (*outString).s[PGCMStringLength-1] = '\0'; 
}


void text_StringToValue (text *value, PGCMString *inString)
/* convert string to text (void) */
{
  int i;
  BOOLEAN terminated;

  i = 0;
  terminated = FALSE;
  while ((!terminated) && (i < MAXTEXTLENGTH))
    {
      value->contents[i] = (*inString).s[i];
      if (value->contents[i] == '\0')
	{
	  terminated = TRUE;
	  /* it's important to write (i+1)! Otherwise the character '\0' will
	     be dropped... */
	  value->curLength = (i+1);
	}
      i++;
    }

  if (!terminated)
    {
      value->contents[MAXTEXTLENGTH - 1] = '\0';
      value->curLength = MAXTEXTLENGTH;
    }
}


/*****************************************************/
/* implementation of user-defined imported functions */
/*****************************************************/

void Text (text *value, PGCMString *inString)
/* implementation of StringToValue, so that 
 * Text(samplestring) can be used instead of value(samplestring, text) */
{
  int i;
  BOOLEAN terminated;

  i = 0;
  terminated = FALSE;
  while ((!terminated) && (i < MAXTEXTLENGTH))
    {
      value->contents[i] = (*inString).s[i];
      if (value->contents[i] == '\0')
	{
	  terminated = TRUE;
	  value->curLength = (i+1);
	}
      i++;
    }

  if (!terminated)
    {
      value->contents[MAXTEXTLENGTH - 1] = '\0';
      value->curLength = MAXTEXTLENGTH;
    }
}


void EmptyText (text *outText)
/* make an empty text (corresponding to "" for strings) */
{
  outText->contents[0] = '\0';
  outText->curLength = 1;
  /* well... it's one character: '\0' */
}


void AmpersandAmpersand (text *outText, text * inText, PGCMString *inString)
/* implementation of the imported function &&, which is an extension of the 
 * built-in string concatenation &. && appends a string to a text */
{
  int i, j;
  BOOLEAN terminated;

  i = 0;
  j = (inText->curLength - 1);
  terminated = FALSE;
  while ((!terminated) && (j < MAXTEXTLENGTH))
    {
      inText->contents[j] = (*inString).s[i];
      if (inText->contents[j] == '\0')
	{
	  terminated = TRUE;
	  inText->curLength = (j+1);
	}
      i++;
      j++;
    }

  if (!terminated)
    {
      inText->contents[MAXTEXTLENGTH - 1] = '\0';
      inText->curLength = MAXTEXTLENGTH;
    }

  *outText = *inText;
}


void ConcatText (text *outText, text *inText1, text *inText2)
/* Concatenation of two texts */
{
  int i, j;
  BOOLEAN terminated;

  i = 0;
  j = (inText1->curLength - 1);
  terminated = FALSE;
  while ((i < (inText2->curLength - 1)) && (j < MAXTEXTLENGTH))
    {
      inText1->contents[j] = inText2->contents[i];
      i++;
      j++;
    }

  if (j == MAXTEXTLENGTH)
    j--;
  inText1->contents[j] = '\0';
  inText1->curLength = (j+1);

  *outText = *inText1;
}


void EqualEqual (BOOLEAN *result, text *inText, PGCMString *inString)
/* implementation of == compares a text with a string 
 * (corresponding to = for strings) */
{
  int i;
  BOOLEAN isequal;

  i = 0;
  isequal = TRUE;
  while ((isequal) && (i < inText->curLength))
    {
      if (inText->contents[i] != (*inString).s[i])
	{
	  isequal = FALSE;
	}
      i++;
    }
  *result = isequal;
}


void IsEqualText (BOOLEAN *result, text *inText1, text *inText2)
/* checks whether two given texts are equal */
{
  int i;
  BOOLEAN isequal;

  if (inText1->curLength == inText2->curLength)
    {
      i = 0;
      isequal = TRUE;
      while ((isequal) && (i < inText1->curLength))
	{
	  if (inText1->contents[i] != inText2->contents[i])
	    {
	      isequal = FALSE;
	    }
	  i++;
	}
    }
  else
    {
      isequal = FALSE;
    }
  *result = isequal;
}

void ajwDummy(BOOLEAN *result) 
{
  fprintf(stderr, "MyDummy Version 2.\n");
  *result = TRUE;
}

