/************************************************************************/
/*									*/
/*    GRAS - A Graph-Oriented Database System for SE Applications	*/
/*    Copyright (C) 1987-1992  Lehrstuhl Informatik III, RWTH Aachen	*/
/*									*/
/*    This library is free software; you can redistribute it and/or	*/
/*    modify it under the terms of the GNU Library General Public	*/
/*    License as published by the Free Software Foundation; either	*/
/*    version 2 of the License, or (at your option) any later version.	*/
/*									*/
/*    This library is distributed in the hope that it will be useful,	*/
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of	*/
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/*    Library General Public License for more details.			*/
/*									*/
/*    You should have received a copy of the GNU Library General Public */
/*    License along with this library; if not, write to the Free	*/
/*    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.*/
/*									*/
/*    Contact Addresses:						*/
/*									*/
/*	Roland Baumann  						*/
/*	Lehrstuhl f"ur Informatik III,					*/
/*	University of Technology Aachen (RWTH Aachen),			*/
/*	Ahornstr. 55,							*/
/*	D-52074 Aachen							*/
/*									*/
/*    Email to								*/
/*									*/
/*	roland@i3.informatik.rwth-aachen.de				*/
/*									*/
/************************************************************************/

/************************************************************************/
/*									*/
/*			A t t r i b u t e d G r a p h			*/
/*									*/
/*----------------------------------------------------------------------*/
/*									*/
/*			 G R A ph - S torage				*/
/*									*/
/*			 Version RWTH- 6.00/11				*/
/*				       | || |				*/
/*				       | || +-denotes the compatibility */
/*				       | ||   of graphs. Each time this */
/*				       | ||   number is incremented the */
/*				       | ||   file format has changed	*/
/*				       | ||				*/
/*				       | |+---denotes minor internal	*/
/*				       | |    changes which do not in-	*/
/*				       | |    fluence your programs	*/
/*				       | |    (e.g. error corrections)	*/
/*				       | |				*/
/*				       | +----denotes minor changes	*/
/*				       |      of the GRAS interface	*/
/*				       |      which may lead to minor	*/
/*				       |      changes in application	*/
/*				       |      programs			*/
/*				       |				*/
/*				       +------denotes the main functio- */
/*					      nal GRAS revisions	*/
/*									*/
/*									*/
/*		      Copyright 1987-1993				*/
/*		   by Lehrstuhl Informatik III,				*/
/*		      RWTH Aachen					*/
/*		      D-52074 Aachen					*/
/*		      Federal Republic of Germany			*/
/*		      Europe						*/
/*									*/
/*----------------------------------------------------------------------*/
/*									*/
/* AUTHORS:								*/
/*   Thomas Brandes, Claus Lewerentz, Andy Schuerr, Ralf Spielmann,	*/
/*   Michael Broekmans, Frank Hoefer, Richard Breuer, and		*/
/*   Thomas Groeger                      				*/
/*									*/
/* RESPONSIBILITY:							*/
/*   Andy Schuerr							*/
/*									*/
/* LAST UPDATE:								*/
/*   July 1993								*/
/*									*/
/*----------------------------------------------------------------------*/
/*									*/
/* This module represents the interface of the IPSEN-"Graphenspeicher". */
/*									*/
/* Exports are:								*/
/*   - operations for managing graph pools				*/
/*   - operations for error recovery					*/
/*   - operations on graphs						*/
/*   - operations for manipulating temporary node sets			*/
/*   - operations for activating event handling mechanisms and		*/
/*   - operations for managing graph schemes & incr. attribute eval.	*/
/*									*/
/* 1. Managing Graph Pools						*/
/* =======================						*/
/*									*/
/*   Terms:								*/
/*									*/
/*   - A GRAPH mainly consists of a set of type-labeled attributed	*/
/*     nodes and type-labeled directed edges between them.		*/
/*   - A GRAPH POOL is a collection of several graphs that are		*/
/*     logically related to each other in one database			*/
/*									*/
/*   Model for multi user processing:					*/
/*									*/
/*     Each user program, to which GRAS is linked, has either read or	*/
/*     write access to a graph at one time. Multiple accesses to one	*/
/*     graph by different user programs are synchronized with the help	*/
/*     of a management graph. Within this graph all currently used	*/
/*     access rights are stored.					*/
/*     These access rights must be returned to GRAS before the user	*/
/*     application terminates. Read or write access to graphs which	*/
/*     have been created by other application programs is not possible	*/
/*     until this graph has been closed for the first time.		*/
/*     The possible opening modes are referred to at 'AGOpenGraph'.	*/
/*									*/
/*   Representation of graphs and graph pools in the file system:	*/
/*									*/
/*     GRAS creates one directory for every graph pool. It is placed in */
/*     a directory specified by an environment variable called "GRAS".	*/
/*     The path name will be constructed from $GRAS, PoolDir, and	*/
/*     PoolName and will be $GRAS/PoolDir/PoolName. The path which is	*/
/*     given by $GRAS/PoolDir must exist before a pool can be created.	*/
/*     Graphs are represented by files in this directory.		*/
/*									*/
/*     A graph A may be represented by another graph B and a chain of	*/
/*     so-called "deltas". These deltas may be used to reconstruct	*/
/*     graph A from the original graph B. Informations about these	*/
/*     reconstructions are stored within the management graph. Local	*/
/*     copies of graphs and all other temporary data produced by GRAS	*/
/*     are stored as files in a directory given by the environment	*/
/*     variable "TMPGRAS".						*/
/*									*/
/*   Environment variables used:					*/
/*									*/
/*   - GRAS	 specifies the directory, in which graph pools are	*/
/*		 stored. There exists no default value for this		*/
/*		 variable. An example for a correct value is		*/
/*		 "/home/project/progress/gras". The directory name must */
/*		 be specified absolutely, not relatively!		*/
/*		 It does not matter whether the name ends with a "/" or */
/*		 not.							*/
/*   - TMPGRAS	 contains the name of the directory, where local graph	*/
/*		 copies are put into. The default is "/usr/tmp".	*/
/*		 It does not matter whether the name ends with a "/" or */
/*		 not.							*/
/*		 Care about the fact, that different users working with */
/*		 GRAS must have different TMPGRAS-directories!		*/
/*   - GRASALLOC gives the amount of heap space that may be allocated	*/
/*		 by GRAS. The unit is 1 kByte. This means that e.g.	*/
/*		 GRASALLOC=100 allows GRAS to allocate 100.000 bytes of */
/*		 heap for internal storage. Note that GRAS will in fact */
/*		 allocate more space, but this is a constant amount and */
/*		 not configurable.					*/
/*		 Appropriate values are 1000 - ...			*/
/*									*/
/* 2. Mechanisms for Recovery						*/
/* ==========================						*/
/*									*/
/*   Besides the possibility to undo all changes on a local graph copy, */
/*   there is the possibility to collect several operations, that make	*/
/*   changes on graphs, into "TRANSACTIONS". These transactions can be	*/
/*   nested to almost arbitrary level. Besides this, check points can	*/
/*   be set, when no transactions are active. Check points mark		*/
/*   consistent graph states, that can be referred to in cases of	*/
/*   errors, program crashes or explicit requests by user applications	*/
/*   (UNDO & REDO).							*/
/*									*/
/*   The following points are characteristic for the concept of trans-	*/
/*   actions:								*/
/*   - one transaction refers to one open graph				*/
/*   - transactions can be nested					*/
/*   - transactions on different graphs are independent from each other */
/*   - only the latest transaction can be interrupted or committed.	*/
/*   - aborting a transaction means undoing exactly all graph changing	*/
/*     operations, that have been executed since the start of the	*/
/*     transaction							*/
/*   - committing a transaction implies of course, that its opera-	*/
/*     tions can be undone by aborting an older transaction (if there	*/
/*     is one...)							*/
/*   - aborting a transaction has no effects on the contents of local	*/
/*     sets!!!								*/
/*									*/
/*   The concept of check points is characterized by the following:	*/
/*   - check points do always refer to one open graph			*/
/*   - check points can only be set, when there is no active trans-	*/
/*     action								*/
/*   - the state of a graph can be reset to any previous check point	*/
/*     (UNDO)								*/
/*   - after UNDO one can return to later checkpoints (REDO), if there	*/
/*     were no changes in between					*/
/*   - two successive checkpoints behave like one if no graph changing	*/
/*     operations were called in between				*/
/*									*/
/*   If a user application crashes, all open graphs are reset to the	*/
/*   state of their latest check point. If there was no check point	*/
/*   defined for an open graph, its state is reset to the one before	*/
/*   the graph has been opened.						*/
/*   If there is an uncontrolled crash (that is if there is no error	*/
/*   message given by GRAS via module Error), the graph state is the	*/
/*   one before opening. The state of the latest check point can be	*/
/*   saved by calling 'AGCheckGraph'.					*/
/*   File system device errors when closing a graph which has been	*/
/*   opened in "WriteOriginal" mode may lead to a corrupt graph state.	*/
/*									*/
/* 3. Graph Changing Operations						*/
/* ============================						*/
/*									*/
/*   Terms:								*/
/*									*/
/*   - TYPE-LABELED NODES represent the objects, that are stored within */
/*     graphs. The type is used for classifying nodes. One node is	*/
/*     identified unambiguously by an internal key, that is created by	*/
/*     GRAS. User suggestions are used, if they don't violate the	*/
/*     above mentioned uniqueness constraint.				*/
/*   - TYPE-LABELED DIRECTED EDGES represent binary relations between	*/
/*     nodes of a graph. Different relations are represented by		*/
/*     different types.							*/
/*   - Every node has several UNSTRUCTURED ATTRIBUTE FIELDS of		*/
/*     "infinite" length, that may be used to store obligate attribute	*/
/*     values.								*/
/*									*/
/* 4. Temporary Sets							*/
/* =================							*/
/*									*/
/*   In order to support storage and processing of query results, GRAS	*/
/*   allows the creation of TEMPORARY SETS, which can store simple	*/
/*   elements (called "SimpleSets") or tuples (called "RelSets").	*/
/*   Note that these sets/relations are main storage data structures 	*/
/*   only and are intended to contain temporarily existing query re-	*/
/*   sults, i.e. they neither permanently stored nor affected by abort	*/
/*   of transactions or undo/redo of graph modifications. There are all	*/
/*   usual operations for sets and binary relations available.		*/
/*									*/
/* 5. Event Handling							*/
/* =================							*/
/*									*/
/*   GRAS supports the activation of user-defined action procedures	*/
/*   depending on certain events.					*/
/*									*/
/*   Terms:								*/
/*									*/
/*   - An ACTION is a user defined procedure.				*/
/*   - An (GRAPH)EVENT can be insertion or deletion of nodes and edges	*/
/*     (each of a certain type), attribute modification, start and end  */
/*     of transactions, setting a checkpoint, undo and redo commands,   */
/*     opening and closing a graph, and user defined events. An event   */
/*     has the following components: the event kind (new node, new edge */
/*     etc.), source and target nodes of operation (used only if the	*/
/*     event affects node(s)), information wether nodes are still       */
/*     existing when event is reported, type information (node type,    */
/*     edge type, attribute number) and a flag wether the reported 	*/
/*     event was triggered by someone else or the notified applications */
/*     itself. These components have different semantics for event      */
/*     specification and reported events (though the same record type   */
/*     is used). 							*/
/*   - A DAEMON is started by the user application for an event and an	*/
/*     action. It activates the action, if a matching event occurs.	*/
/*     Occuring events match, if the TNo (TypeNumber) and S/TNode para-	*/
/*     meters of the event match that of the activated daemon. 	It is   */
/*     possible to specify, that an activated daemon matches every TNo/	*/
/*     S/TNode by setting the appropiate field to 0 (wildcard) when the	*/
/*     daemon is activated. Furthermore the application can decide 	*/
/*     wether events from all clients of a graph should be reported or  */
/*     not, by setting the Boolean self to false or true respectively.  */
/*     If the application is only interested in events for which the    */
/*     source node or target node still exist when the event is re-    	*/
/*     ported it will set SNodeExists or TNodeExists field (or both)	*/
/*     of the GraphEvent-Record to TRUE, otherwise all events will be	*/
/*     reported. In short, when specifying an event for a daemon, the   */
/*     components of a GraphEvent-Record have the following meaning:	*/
/*        kind: what kind of operation triggers the event		*/
/*        SNode: (Source) node number for node, edge and attribute      */
/*               operations. 0 is matches all node numbers		*/
/*        SNodeExists: TRUE -> report event only if (source) node still */
/*		       exists. FALSE -> report all events.		*/
/*        TNode: Target node number for edge operations. 0 is wildcard. */
/*        TNodeExists: TRUE -> report event only if target node still 	*/
/*		       exists. FALSE -> report all events.		*/
/*	  TNo: Type number for nodes, edges, attributes. 0 matches all	*/
/*	       types.							*/
/*	  Self: TRUE -> report only if application triggered the event. */
/*		FALSE -> report events triggered by other clients.	*/
/*     Note that is does not make sense to specify SNodeExists=TRUE for */
/*     delete-node-events. This will be ignored and delete node events  */
/*     are always reported.						*/
/*   - ACTIVATED ACTIONS are performed after termination of the		*/
/*     initiating GRAS-procedure in a sequence, that depends on their	*/
/*     priority. (Note: The purpose for open and close graph events is  */
/*     for clients to be able to notify if other clients operate on the */
/*     same graph.) Depending on the kind of event that occured, the	*/
/*     components of the GraphEvent-record given as argument to the     */
/*     invoked action are                                               */
/*                                                                      */
/*     EventKind       | TNo               | SNode         | TNode 	*/
/*     ---------------------------------------------------------------  */
/*     NewNode         | type of the new / | number of the | -   	*/ 
/*     DeleteNode      | deleted node      | new/del. node | -		*/
/*                     |		   |		   | 		*/
/*     NewEdge	       | type of the new / | source        | target 	*/
/*     DeleteEdge      | deleted edge	   | node numer    | node number*/
/*	    	       |		   |		   |	 	*/
/*     ModifyAttribute | attr. number	   | node number   | -		*/
/*		       |		   |		   |		*/
/*     TransactionStart| -		   | -		   | -		*/
/*     TransactionEnd  | = 0 -> commit, # 0 -> abort	   | -		*/
/*     SetCheckpoint   | -		   | -		   | -		*/
/*     Undo	       | number of un/re-  | -		   | -		*/
/*     Redo	       | done steps	   | -		   | -		*/
/*     OpenGraph       | -		   | -		   | -		*/
/*     CloseGraph      | -		   | -		   | -		*/
/*     UserDefined     | p1		   | p2		   | p3		*/
/*		       | of call to AGRaiseEvent	   |	 	*/
/*									*/
/*   The last component of a GraphEvent record, the Boolean Self ,      */
/*   states whether this event was triggered by the client who now      */
/*   receives the event or not. SnodeExists and TNodeExists state 	*/
/*   wether the respective nodes still exist at event report time.	*/
/*   The execution of an action procedure can be delayed in order to	*/
/*   wait e.g. until complex graph manipulations have been completed.	*/
/*   Nesting of delays is supported. Delayed but active actions will be */
/*   updated when corresponding nodes are deleted in the meantime. 	*/
/*   Either their S/TNodeExists-Falgs will be set to FALSE or the 	*/
/*   action will be cancelled, depending on the value of the S/TNode-	*/
/*   Exists-Fields of triggering daemon.				*/
/*   During the execution of an action procedure other actions are	*/
/*   blocked.								*/
/*   The following GRAS-procedures may only be called, when action	*/
/*   processing is allowed (to avoid cancellation of blocked actions):	*/
/*   - AGCopyGraph, AGCloseGraph, AGDeleteGraph, AGDeleteGraphPool,	*/
/*     AGStartTransaction, AGCommitTransaction, AGAbortTransaction,	*/
/*     AGKillDaemon, (AGCheckGraph), and user defined actions.		*/
/*									*/
/* 6. GRAS Error Messages						*/
/* ======================						*/
/*									*/
/*   There are three different kinds of error messages:			*/
/*									*/
/*   - error messages in GRAS, that are initiated by application errors */
/*     (happy searching...). Some possible reasons for these errors are:*/
/*									*/
/*     a) Errors which result of incorrect use of the GRAS interface.	*/
/*	  These errors are mentioned within the procedure documentary	*/
/*	  comments.							*/
/*     b) Application run time errors revealed by the MODULA2 system	*/
/*									*/
/*     In both cases GRAS is terminated correctly. That means, that	*/
/*     local copies of all open graphs will be created. Then these	*/
/*     graphs will be reset into consistent state (see Error Recovery)	*/
/*     and closed.							*/
/*     Additionally a CORE DUMP of the application is created and the	*/
/*     user program is halted.						*/
/*									*/
/*   - internal error messages can be recognized by being rather	*/
/*     cryptic. In this case GRAS tries to perform the actions that	*/
/*     were mentioned above (the GRAS programmers' team apologizes).	*/
/*									*/
/*   - errors concerning the UNIX file system. Possible reasons are:	*/
/*     a) the disk is full						*/
/*     b) undefined environment variable GRAS				*/
/*     c) non existent directory TMPGRAS				*/
/*     d) illegal user access to UNIX directories or files.		*/
/*     e) crash of the NFS						*/
/*     In case of one of these errors all write-opened graphs are	*/
/*     locked and must be unlocked by calling 'AGCheckGraph' or		*/
/*     'AGReorganizeGraphPool' (see Error Recovery) before any access	*/
/*     is possible again.						*/
/*									*/
/* 7. Scheme Management							*/
/* ====================							*/
/*									*/
/*   GRAS supports graph schemes. They may be declared using several	*/
/*   procedures whose headers appear after the comment entry "Scheme	*/
/*   Management". By means of these graph schemes GRAS is able to	*/
/*   guarantee a graph's consistency with its graph scheme (preventing	*/
/*   e.g. assignments to undeclared attributes or creation of edges bet-*/
/*   ween nodes which do not have correct types. Besides, graph 	*/
/*   schemes contain all information necessary for the incremental eva- */
/*   luation of derived attributes (using the well-known two-phase lazy */
/*   attribute evaluation algorithm). Therefore, our graph schemes	*/
/*   contain the following types of information:			*/
/*									*/
/*   - NODE CLASSES which are sets of node types. Properties defined	*/
/*     for one node class are inherited to all subclasses of this node	*/
/*     class and finally to all node types (and nodes having these	*/
/*     types).								*/
/*									*/
/*   - NODE TYPES which are used as declarations of node types and are	*/
/*     considered to be instances of node classes. All nodes which are 	*/
/*     instances of the same node type do have			 	*/
/*     the same attribute evaluation procedures and dependencies.	*/
/*     Sometimes the term NODE LABEL is used synonymously.		*/
/*									*/
/*   - EDGE TYPES which are nothing else but declarations of edge 	*/
/*     types. The source and target class of edges of one type are 	*/
/*     fixed. Thus, every source/target of a certain edge must be an 	*/
/*     instance of a node type of the edge type's source/target class.	*/
/*     There are two categories of edge types with edges of the first	*/
/*     category sharing the common property that targets of these edges */
/*     are existent-dependent on their sources.				*/
/*     Sometimes the term EDGE LABEL is used synonymously.		*/
/*									*/
/*   - ATTRIBUTES within GRAS belong to different categories:		*/
/*									*/
/*     AttributeKind							*/
/*     -------------							*/
/*     All attributes are INTRINSIC or META or DERIVED or DUMMY.	*/
/*     Intrinsic ones receive their values through explicit assignment	*/
/*     only (with exception of an automatically computed initial value)	*/
/*     whereas derived attributes depend on the values of other attri-  */
/*     butes and will be computed within the GRAS system (if GRAS has	*/
/*     access to all application-specific evaluation functions).	*/
/*     Meta attributes are a special variant of intrinsic attributes	*/
/*     which receive a node type dependent value (node type attribute).	*/
/*     And dummy attributes are a special class of derived attributes   */
/*     which do not possess a real value but are either defined or un-  */
/*     defined. These attributes are used for propagating graph changes */
/*     along arbitrarily long paths in the graph, i.e. for defining     */
/*     attribute dependencies between two attributes which are not in   */
/*     the one context of each other.					*/
/*									*/
/*     IndexProperties							*/
/*     ---------------							*/
/*     GRAS supports associative accesses to nodes/attributes by distin-*/
/*     guishing four different categories of attributes:		*/
/*									*/
/*     - NORMAL attributes (with a maximum length of 251*65535=16449285	*/
/*	 bytes each) for which associative access is not supported	*/
/*     - INDEX attributes (with a maximum length of 240 bytes) for	*/
/*	 which by means of an index structure GRAS is able to 		*/
/*	 efficiently answer queries like: "Find all nodes whose index 	*/
/*	 attribute A has the value V"					*/
/*     - KEY attributes (with a maximum length of 240 bytes) which	*/
/*	 are special index attributes with the additional property that	*/
/*	 the above mentioned query always delivers at most one node.	*/
/*									*/
/*   The "manual" construction of graph schemes by means of the GRAS	*/
/*   system's scheme procedures (explained below) is very tedious. But	*/
/*   all those GRAS applications which do not have to exploit the GRAS	*/
/*   system's capabilities for consistency checking and incremental	*/
/*   attribute evaluation are not forced to declare graph schemes (in	*/
/*   this case the error check mode must be set to "inactive").		*/
/*   Furthermore, an integrated environment for the specification of	*/
/*   graph schemes, graph rewrite rules etc. is under development.	*/
/*   A first version of this environment including a syntax directed	*/
/*   editor, an incrementally working consistency checker, a graph	*/
/*   browser, a graph scheme compiler etc. is near-by completion.	*/
/*									*/
/*									*/
/* Features & Bugs		 					*/
/* ===============							*/
/*									*/
/*   - The following things might cause a decrease of efficiency:	*/
/*     a) oscillating accesses to nodes, that are not physically	*/
/*	  neighbored (therefore logically neighbored nodes are to be	*/
/*	  stored as neighbors)						*/
/*     b) too many nodes within one graph (but don't worry about up to	*/
/*	  30,000 nodes...)						*/
/*     c) too many edges with the same type, that come in or go out of  */
/*	  one single node (don't worry about up to 1,000 edges...)	*/
/*	  The current limit is about 5000 edges of the same type and	*/
/*	  the same direction for one node.				*/
/*   - Writing operations on a graph, that was opened for read-only	*/
/*     access are not forbidden, but ignored, when the graph is finally */
/*     closed!								*/
/*   - Deactivating the error check mode increases the GRAS system's	*/
/*     performance and allows for the manipulation of graphs without	*/
/*     the declaration of graph schemes (this might be the default for	*/
/*     the development of efficient implementations of abstract data	*/
/*     types).								*/
/*   - The 'UseMembership' flag in 'DeclareNodeType' may not be used 	*/
/*     if more than 5000 nodes of this type will be created (and avoid  */
/*     using this flag whenever possible for efficiency reasons).	*/
/*   - File system device errors when closing a graph which has been	*/
/*     opened in "WriteOriginal" mode may lead to a corrupt graph state.*/
/*   - Inheritance of attributes does not work in the case of extending */
/*     already existing/committed graph schemes.			*/ 
/*									*/
/*----------------------------------------------------------------------*/

/*
 *
 * C-interface to the remote GRAS-system.
 */
#if !defined (_RGRAS_)

#define _RGRAS_

#ifdef __cplusplus
extern "C" {
#endif

#include "rgglobal.h"

/*!!! REI : hinzugefuegt aus csci/param.h*/
extern void
RealCloseGraph(GraphNumber graph,
	       BOOLEAN     force);


extern void
AGLogin(const ApplicationName name);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Initialization of the GRAS system. Name is used	*/
		/*   in status reports.					*/
		/*							*/
		/********************************************************/

extern void
AGLogout(void);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Logging out the GRAS system			*/
		/*							*/
		/********************************************************/

extern void
AGTimerOn(CARDINAL interval);
                /********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Initialization of a timer that interrupts the 	*/
                /*   application regularly to check for incoming        */
		/*   messages. The parameter interval specifies the     */
                /*   time interval in ms between two interruptions.     */
		/*   A good value might be 20.  			*/
		/*							*/
		/********************************************************/

extern void
AGTimerOff(void);
                /********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
                /*   Turn timer off.                             	*/
		/*							*/
		/********************************************************/

/************************************************************************/
/*									*/
/*		  Operations for managing graph pools			*/
/*									*/
/************************************************************************/

extern void
AGOpenGraphPool(const PoolName Pool,
                      GraphPoolMode mode,    
                      TStatus  *Status);
 		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Creation of a graph pool which can hold several	*/
		/*   graphs. Also: Inform ganimed that pool is in use.	*/
		/*							*/
		/* Errors Returned:					*/
		/*							*/
		/*   AlreadyExistent: pool already exists		*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SYNTAX E. : illegal pool name			*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGCloseGraphPool(const PoolName Pool);
 		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Inform ganimed that pool is no longer in use.	*/
		/*							*/
		/********************************************************/

extern void
AGDeleteGraphPool(const PoolName Pool,
		  TStatus  *Status);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes a pool if there are no graphs left in it.	*/
		/*							*/
		/* Errors Returned:					*/
		/*							*/
		/*   NotExistent: pool does not exist			*/
		/*   NotEmpty	: there is at least one graph left in	*/
		/*		  the pool				*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   BUSY      : too many applications working with the */
		/*		 graph pool				*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGReorganizeGraphPool(const PoolName Pool,
		      TStatus  *Status);
                /********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Reorganizes the graph pool by removing all unused	*/
		/*   information from the management graph. Besides,	*/
		/*   all unnecessary GRAS files are deleted. Deltas are	*/
		/*   packed and combined if possible. Furthermore all	*/
		/*   existing locks on graphs are released (which are	*/
		/*   due to application or system crashes) and recovery	*/
		/*   is initiated for all previously locked graphs	*/
		/*   based on information in TMPGRAS (see also AGCheck-	*/
		/*   Graph).						*/
		/*							*/
		/* Warning:						*/
		/*							*/
		/*   This procedure may only be called by privileged	*/
		/*   application modules, because any access to the	*/
		/*   graph pool is locked until AGReorganizeGraphPool	*/
		/*   is terminated.					*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   AGReorganizeGraphPool may take a rather long time	*/
		/*   to do its job!					*/
		/*							*/
		/* Errors Returned:					*/
		/*							*/
		/*   AlreadyExistent: pool already exists		*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SYNTAX E. : illegal pool name			*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGRecoverGraphPool(const PoolName Pool,
		      TStatus  *Status);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
                /*   are due to application or system crashes) and      */
                /*   recovery is initiated for all previously locked    */
                /*   graphs based on information in TMPGRAS.            */
		/*							*/
		/* Warning:						*/
		/*							*/
		/* Note:						*/
		/*							*/
		/* Errors Returned:					*/
		/*							*/
		/*   AlreadyExistent: pool already exists		*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SYNTAX E. : illegal pool name			*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGResetGraphPool(const PoolName Pool,
		      TStatus  *Status);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   All existing locks on graphs are released (which 	*/
                /*   are due to application or system crashes). 	*/
		/* Warning:						*/
		/*							*/
		/* Note:						*/
		/*							*/
		/* Errors Returned:					*/
		/*							*/
		/*   AlreadyExistent: pool already exists		*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SYNTAX E. : illegal pool name			*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/



extern void
AGGetGraphsInGraphPool(const PoolName Pool,
		       BOOLEAN  *Existent);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Tests, whether the pool exists. If so, a list of	*/
		/*   all external graph numbers of the graphs in this	*/
		/*   pool is created. It can be referred to with the	*/
		/*   procedure "AGGiveNextGraphInGraphPool".		*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SYNTAX E. : illegal pool name			*/
		/*   BUSY      : too many applications working with the */
		/*		 graph pool				*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGGiveNextGraphInGraphPool(ExternNumber *ExtNumber,
			   BOOLEAN      *Existent);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   The result of this function is the next external	*/
		/*   graph number in the list created by the procedure	*/
		/*   "AGGetGraphsInGraphPool".				*/
		/*							*/
		/********************************************************/

extern void
AGCopyGraph(GraphNumber SourceGraph,
	    const PoolName    TargetPool,
	    const GraphName   TargetGrName,
	    TStatus     *Status);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Creates a copy of an open graph and places it in	*/
		/*   pool 'TargetPool'. 'TargetGraph' can be used	*/
		/*   directly and doesn't need to be reconstructed with */
		/*   a delta.						*/
		/*							*/
		/* Errors Returned:					*/
		/*							*/
		/*   NotExistent    : 'TargetPool' doesn't exist	*/
		/*   AlreadyExistent: graph already exists		*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SYNTAX E. : illegal graph or pool name		*/
		/*   IS CLOSED : 'SourceGraph' not opened		*/
		/*   PENDING T.: still a transaction running on source	*/
		/*		 graph					*/
		/*   BUSY      : too many applications running on the	*/
		/*		 pool					*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGDeltaCopyGraph(const PoolName  Pool,
		 const GraphName SourceGrName,
		 const GraphName TargetGrName,
		 BOOLEAN   Forward,
		 TStatus   *Status);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Copies the closed graph "SourceGrName" from the	*/
		/*   pool "Pool" logically to "TargetGrName" in the same*/
		/*   pool. If "Forward" is set to TRUE, the copy is	*/
		/*   performed with a forward delta, else with a back-	*/
		/*   ward delta.					*/
		/*							*/
		/* Note:						*/
		/*   Take care of the fact, that AGDeltaCopyGraph needs */
		/*   a closed source graph whereas AGCopyGraph needs an */
		/*   open one!						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SYNTAX E. : illegal graph or pool name		*/
		/*   BUSY      : too many applications working with the */
		/*		 graph pool				*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGGetFirstGroupInfo(const PoolName  Pool,
		    const GraphName GrName,
		    BOOLEAN   *OneGroup,
		    BOOLEAN   *Existent,
		    GroupName Group,
		    GroupType *LockMode);
                /********************************************************/
                /*                                                      */
                /* Purpose:                                             */
                /*                                                      */
                /*   Tests wether the graph is opened AND an access     */
                /*   group exists. Then the parameter OneGroup is       */
                /*   TRUE, if the graph might be accessed only          */
                /*   by members in exactly this group (i.e. no other    */
                /*   groups can exist at the same time). The identi-    */
                /*   fication of the group is returned in Group.        */
                /*   The lock duration to the graph for all group       */
		/*   members is returned in LockMode.			*/
                /*                                                      */
                /* Note:                                                */
                /*                                                      */
                /*   If the parameter Existent is FALSE just as it was  */
                /*   in the call to AGGetWriterGroupInfo then the graph */
                /*   is not opened.                                     */
                /*   If the parameter OneGroup is TRUE, a subsequent    */
                /*   call to AGGetNextReaderGroupInfo always yields     */
                /*   Existent = FALSE.                                  */
                /*                                                      */
                /********************************************************/

extern void
AGGetNextGroupInfo(BOOLEAN   *Existent,
		   GroupName Group,
		   GroupType *LockMode);
                /********************************************************/
                /*                                                      */
                /* Purpose:                                             */
                /*                                                      */
                /*   Returns information on the next group if existent. */
                /*                                                      */
                /********************************************************/ 

extern void
AGGetInfoByGrNo(GraphNumber  Graph,
		BOOLEAN      *IsOpen,
		PoolName     Pool,
		GraphName    GrName,
		ExternNumber *ExtNumber,
		GraphType    *GrType,
		GroupName    Group,
		GroupType    *LockMode,
		BOOLEAN      *SingleWriter);
		/********************************************************/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "GrNumber" not opened		*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGGetInfoByExtNo(const PoolName     Pool,
		 ExternNumber ExtNumber,
		 BOOLEAN      *Existent,
		 GraphName    GrName,
		 GraphType    *GrType,
		 BOOLEAN      *IsOpen,
		 GraphNumber  *Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Checks, whether a graph with the external number	*/
		/*   "ExtNumber" exists in the pool. If so, the		*/
		/*   following information is returned:			*/
		/*   - the graph type					*/
		/*   - its name						*/
		/*   - whether the graph is opened by the current	*/
		/*     application and if so,				*/
		/*     - its temporary graph number			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SYNTAX E. : illegal graph or pool name		*/
		/*   BUSY      : too many applications working with the */
		/*		 graph pool				*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGGetInfoByName(const PoolName     Pool,
		const GraphName    GrName,
		BOOLEAN	     *Existent,
		GraphType    *GrType,
		ExternNumber *ExtNumber,
		BOOLEAN	     *IsOpen,
		GraphNumber  *Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Checks, whether a graph with the specified name	*/
		/*   "GrName" exists in the pool. If so, the following	*/
		/*   information is returned:				*/
		/*   - the graph type					*/
		/*   - its external number				*/
		/*   - whether the graph is opened by the current	*/
		/*     application and if so,				*/
		/*     - its temporary graph number			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SYNTAX E. : illegal graph or pool name		*/
		/*   BUSY      : too many applications working with the */
		/*		 graph pool				*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGOpenGraph(const PoolName  Pool,
	    const GraphName GrName,
	    GraphType	    *GrType,
	    GraphMode	    OpenMode,
	    const GroupName GpName,
	    GroupType       LockType,
	    GroupMode	    GpMode,
	    BOOLEAN	    OneGroup,
	    BOOLEAN	    NoWait,
	    TStatus	    *Status,
	    ExternNumber    *ExtNumber,
	    GraphNumber	    *Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
                /*   Creates or opens a graph of GraphType GrType with  */
                /*   a call to AGCreateTypedGraph or AGOpenGraph. If    */
                /*   created then closes the graph and uses AGOpenGraph */
                /*   with the appropriate parameters for opening the    */
                /*   graph again.                                       */
                /* Functionality:                                       */
                /*                                                      */
                /*   Opens a graph referred to by Pool and GrName with  */
                /*   access GroupMode and AccessRight. If OneGroup is	*/
                /*   TRUE the graph is locked against any further       */
                /*   attempt to create new access groups via calls of   */
                /*   AGSharedOpenGraph. If group access is SameReader   */
                /*   the GroupID identifies the group in which the      */
                /*   application will be a member.                      */
                /*   The GroupID is ignored in the cases the client     */
                /*   shares a writer group.                             */
                /*   If NoWait is TRUE the function returns immediately */
                /*   not waiting for satisfying its request if it       */
                /*   cannot be handled instantaneously.                 */
                /*                                                      */
                /* Note:                                                */
                /*                                                      */
                /*   Because of the nature of the new access model it   */
                /*   can take a (very) long time to get a response for  */
                /*   the opening request particularly if the desired    */
                /*   access right is Exclusive. Therefore it is wise    */
                /*   to use NoWait = TRUE.                              */
                /*                                                      */
                /* Errors Returned:                                     */
                /*   GroupNotExistent     : group does not exist        */
                /*   GroupAlreadyExistent : group already existent      */
                /*   DifferentAccessRights: specified access right      */
                /*                          differs from that of group  */
                /*   TooManyGroupMember   : no more members possible    */
                /*   TooManyGroups        : no more groups available    */
                /*   Locked               : graph is locked against any */
                /*                          attempt to share other than */
                /*                          the only existing group     */
                /*   RequestWouldbeDelayed: request would be success-   */
                /*                          full, if NoWait = FALSE.    */
                /*                                                      */
                /* Note:                                                */
                /*                                                      */
                /*   In addition to the error codes mentioned above     */
                /*   errors are returned which occur due to the use     */
                /*   of the standard open function AGOpenGraph.         */
                /*                                                      */
                /* Important:                                           */
                /*                                                      */
                /*   'Status' can hold a different value in addition    */
                /*   to the 'NoError' value that indicates a            */
                /*   successful but considerable restricted use of      */
                /*   further operations on the now opened graph. This   */
                /*   value is:                                          */
                /*   NoWriteAccess       : graph changing operations    */
                /*                         are not allowed              */
                /*                                                      */
		/********************************************************/

extern void
AGCloseGraph(GraphNumber Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Closes a graph and resets the corresponding access */
		/*   locks. This procedure may only be called at a time */
		/*   when all transactions concerning the graph are	*/
		/*   committed.If the graph has been opened in SaveCopy */
		/*   or WriteOrg mode, the parameter "Save" specifies	*/
		/*   whether the changes made since opening time will	*/
		/*   be saved or not.					*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   System crashes during execution of operations may	*/
		/*   lead to corruption in case of (and only of) the	*/
		/*   opening mode WriteOrg and Save=TRUE.		*/
		/*							*/
		/*   A graph that was created and not closed before,	*/
		/*   will not be existing anymore, if it is closed	*/
		/*   with Save=FALSE!					*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   FORBIDDEN : there are still locked active actions	*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   BUSY      : too many applications working with the */
		/*		 graph pool				*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGDeleteGraph(const PoolName  Pool,
	      const GraphName GrName,
	      BOOLEAN   Force,
	      TStatus   *Status);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   The closed graph "GrName" in the GraphPool "Pool"	*/
		/*   will be deleted.				  	*/
		/*							*/
		/*   If "Force" is TRUE,			  	*/
		/*   this procedure tries to reset all write locks, if	*/
		/*   there are any. If this is not possible, "Status"	*/
		/*   is set to "Locked".				*/
		/*   If there is a lock on the management graph, one	*/
		/*   should try to copy all graphs from the pool into	*/
		/*   a new GraphPool (...if this is still possible with */
		/*   the corrupted management graph). Therefore one	*/
		/*   needs to reset all locks with the help of the	*/
		/*   operating system.					*/
		/*							*/
		/* Errors Returned:					*/
		/*							*/
		/*   NotExistent: graph or pool does not exist		*/
		/*   StillOpen  : the graph is still opened by          */
		/*                an application program      		*/
		/*   Locked	: graph is locked by a different	*/
		/*		  application program			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   BUSY      : too many applications working with the */
		/*		 graph pool				*/
		/*   NOT LOGGED: application has not called AGLogin	*/
		/*		 before trying this operation		*/
		/*							*/
		/********************************************************/

extern void
AGRequestSingleWriter(GraphNumber Graph,
		      BOOLEAN	  Wait,
		      TStatus	  *Status);
                /********************************************************/
                /*                                                      */
                /* Purpose:                                             */
                /*                                                      */
                /*   Requests to become the only group member who is	*/
		/*   allowed to modify the graph. If this right is	*/
		/*   already given to another member, either (Wait is	*/
		/*   TRUE) wait	until this right is granted or return	*/
		/*   immediately.					*/
                /*                                                      */
		/* Errors returnd:					*/
                /*                                                      */
                /*   WriterExistent	    : the right is given to	*/
                /*                            another client            */
                /*                                                      */
                /********************************************************/

extern void
AGReleaseSingleWriter(GraphNumber Graph);
                /********************************************************/
                /*                                                      */
                /* Purpose:                                             */
                /*                                                      */
                /*   Releases the exclusive write access within this 	*/
		/*   group gained by AGRequestSingleWriter.		*/
                /*                                                      */
		/* Errors returnd:					*/
                /*                                                      */
		/*   IS CLOSED : graph "Graph" not opened		*/
                /*                                                      */
                /********************************************************/

/************************************************************************/
/*									*/
/*	      Operations for Primitive Recovery Mechanisms		*/
/*									*/
/************************************************************************/

extern void
AGStartToplevelTransaction(GraphNumber Graph,
			   BOOLEAN     NoWait,
			   TStatus     *Status);
                /********************************************************/
                /*                                                      */
                /* Purpose:                                             */
                /*                                                      */
                /*   Starts a transaction for the specified graph.      */
                /*   This and the operations AGCommitTransaction        */
                /*   and AGRollbackTransaction are only permitted for   */
                /*   graphs opened with access right 'transactional'    */
                /*   or 'exclusive'. With 'transactional' access        */
                /*   at least one transaction must be started before    */
                /*   any other operation on this graph takes place.     */
                /*   While running a transaction the graph is protected */
                /*   against access from other clients.                 */
                /*   If the parameter NoWait is TRUE the function       */
                /*   returns immediatly after trying to start a trans-  */
                /*   action. Otherwise control is returned to the       */
                /*   caller after that the transaction was started      */
                /*   sucessfully. The function AGStartTransaction can   */
                /*   be used furthermore, but the return from this call */
                /*   is always delayed until successful termination.    */
                /*                                                      */
                /* Errors Returned:                                     */
                /*                                                      */
                /*   StillPendingTransaction: the graph is accessed by  */
                /*                            another client            */
                /*                                                      */
                /********************************************************/

extern void
AGStartTransaction(GraphNumber Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Starts a new transaction. This transaction can	*/
		/*   either be committed normally (then it has no       */
		/*   effects on the state of the graph) or be aborted	*/
		/*   (that means, that the graph is reset to the state	*/
		/*   when AGStartTransaction was called).		*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   All restrictions and effects of this procedure are */
		/*   mentioned in the module comments.			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   FORBIDDEN : there are still locked active actions	*/
		/*		 or pending transaction			*/
		/*							*/
		/********************************************************/

extern void
AGCommitToplevelTransaction(GraphNumber Graph);
                /********************************************************/
                /*                                                      */
                /* Purpose:                                             */
                /*                                                      */
                /*   Commits all transactions that were started on the  */
                /*   specified graph.                                   */
                /*                                                      */
                /********************************************************/

extern void
AGCommitTransaction(GraphNumber Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Commits a transaction that was started by calling	*/
		/*   AGStartTransaction in the normal way. That means,	*/
		/*   that all graph changing operations performed	*/
		/*   during the transaction take place in the graph.	*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   FORBIDDEN : there are still locked active actions	*/
		/*		 or pending transaction			*/
		/*   NO TRANS. : there is no more active transaction	*/
		/*							*/
		/********************************************************/

extern void
AGAbortToplevelTransaction(GraphNumber Graph);
                /********************************************************/
                /*                                                      */
                /* Purpose:                                             */
                /*                                                      */
                /*   Resets the graph to the state right before         */
                /*   initiating the first transaction.                  */
                /*                                                      */
                /********************************************************/

extern void
AGAbortTransaction(GraphNumber Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Aborts a transaction that was started by calling	*/
		/*   AGStartTransaction. The effect is, that the graph	*/
		/*   is reset into the state at the time the trans-	*/
		/*   action was started.				*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   FORBIDDEN : there are still locked active actions	*/
		/*		 or pending transaction			*/
		/*   NO TRANS. : there is no more active transaction	*/
		/*							*/
		/********************************************************/

extern void
AGPendingTransactions(GraphNumber Graph, CARDINAL *depth);
		/********************************************************/
		/*                                                      */
		/* Purpose:                                             */
		/*                                                      */
		/*   Returns the number of nested transaction currently */
		/*   active.                                            */
		/*                                                      */
		/* Error Messages:                                      */
		/*                                                      */
		/*   IS CLOSED : graph "Graph" not opened               */
		/*                                                      */
		/********************************************************/

/************************************************************************/
/*									*/
/*	Operations for Requests and Modifications on Single Graphs	*/
/*									*/
/* Note:								*/
/*									*/
/*   The effects of these operations can be reset by aborting trans-	*/
/*   actions (but this does not change the state of local sets!)	*/
/*									*/
/************************************************************************/

extern void
AGCreateNode(GraphNumber Graph,
	     TypeNumber	 NType,
	     NodeNumber	 Environment,
	     NodeNumber	 *NodeNr);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Creates a new node in "Graph". This node gets the	*/
		/*   type "NType". The parameter "Environment" may	*/
		/*   specify another node, which represents a logical	*/
		/*   neighbor to the current one. "NodeNr" passes the	*/
		/*   number of the new node to the application. It is	*/
		/*   set to "Environment", if that number is not yet	*/
		/*   used.						*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   The parameter "Environment" is internally used to	*/
		/*   place the nodes onto pages in order to minimize	*/
		/*   disk accesses. Therefore logically neighbored	*/
		/*   nodes should always be stored as physically neigh- */
		/*   bored ones.					*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*							*/
		/********************************************************/

extern void
AGCreateEdgeAndNode(GraphNumber Graph,
		    NodeNumber  SourceNode,
		    TypeNumber  EType,
		    TypeNumber  NType,
		    NodeNumber  *NodeNr);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Inserts a new edge from node "SourceNode" to a	*/
		/*   node of type "NType". The edge gets the type	*/
		/*   "EType". "NodeNr" returns the number of the new	*/
		/*   node. The new node is placed in physical neigh-	*/
		/*   borhood to the existing one.			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   NOT FOUND : the node "SourceNode" does not exist	*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern void
AGDeleteNodeAndEdges(GraphNumber Graph,
		     NodeNumber	 Node);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes the specified node with all its incoming	*/
		/*   and outgoing edges from the graph.			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   NOT FOUND : the node "Node" does not exist		*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*							*/
		/********************************************************/

extern void
AGGetUniqueNodeId(GraphNumber Graph,
		  UniqueID    *UID);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   This procedure computes a name which is unique for */
		/*   the whole graph pool. The name consists of two	*/
		/*   ordinal numbers which TOGETHER (as a tuple) ensure */
		/*   uniqueness.					*/
		/*							*/
		/********************************************************/

extern void
AGShowAllNodesWithIndex(GraphNumber     Graph,
			AttributeNumber AttNo,
			CARDINAL        Length,
			const BYTE      *Value,
			SimpleSet       *SetOfNodes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns a set of nodes the given attribute value 	*/
		/*   is associated with.				*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   This procedure may not be used within evaluation-  */
                /*   functions of attributes.                           */
                /*                                                      */
		/*   Furthermore, in the case of a (boolean) constraint	*/
		/*   attribute this query is only supported for "Value" */
		/*   equal to "Violated" = FALSE (with "Length" = 1) in */
		/*   order to avoid creation of a large useless index	*/
		/*   for all nodes which respect a certain constraint.  */
                /*                                                      */
		/*   "Value" is VAR parameter for efficiency reasons    */
		/*   only.						*/
		/*   							*/
		/*   Use "AGSetRemoveAnyElement" to access all nodes	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED     : graph "Graph" not opened		*/
                /*   INVALID CYCLE : use in an evaluationfunction       */
		/*   UNKNOWN   : set "SetOfNodes" has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGShowNodeWithKey(GraphNumber	  Graph,
		  AttributeNumber AttNo,
		  CARDINAL	  Length,
		  const BYTE	  *Value,
		  BOOLEAN	  *Exist,
		  NodeNumber	  *Node);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns the number of the node, the given attribute*/
		/*   value is associated with.				*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   This procedure may not be used within evaluation-  */
                /*   functions of attributes.                           */
                /*                                                      */
		/*   Furthermore, in the case of a (boolean) constraint	*/
		/*   attribute this query is only supported for "Value" */
		/*   equal to "Violated" = FALSE (with "Length" = 1) in */
		/*   order to avoid creation of a large useless index	*/
		/*   for all nodes which respect a certain constraint.  */
                /*                                                      */
		/*   "Value" is VAR parameter for efficiency reasons    */
		/*   only.						*/
                /*                                                      */
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED     : graph "Graph" not opened		*/
                /*   INVALID CYCLE : use in an evaluationfunction       */
		/*   EXISTENT  : there is mode than one node with the	*/
		/*		 key attribute (AttNo,Name), which is a	*/
		/*		 violation of the "Key" property	*/
		/*							*/
		/********************************************************/

extern void
AGShowTypeOfNode(GraphNumber Graph,
		 NodeNumber  Node,
		 TypeNumber  *NType);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns the type of a given node. If the node does */
		/*   not exist, NType is set to UndefinedTypeNumber=0.  */
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*							*/
		/********************************************************/

extern void
AGCreateEdge(GraphNumber Graph,
	     NodeNumber	 SourceNode,
	     NodeNumber	 TargetNode,
	     TypeNumber	 EType);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Inserts a new directed edge from node "SourceNode" */
		/*   to node "TargetNode". This edge gets the type	*/
		/*   "EType".						*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   Edges of the composite category can only be	*/
		/*   created s incoming ones, when the node is created. */
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   NOT FOUND : "SourceNode" or "TargetNode" does not	*/
		/*		 exist					*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern void
AGDeleteEdge(GraphNumber Graph,
	     NodeNumber	 SourceNode,
	     NodeNumber	 TargetNode,
	     TypeNumber	 EType);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes the edge of type "EType" between the	*/
		/*   nodes "SourceNode" and "TargetNode".		*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   Edges of the composite category can only be	*/
		/*   deleted together with their corresponding nodes.	*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SIZE ERROR: there are too many open graphs		*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern BOOLEAN
AGEdgeExists(GraphNumber Graph,
	     NodeNumber	 SourceNode,
	     NodeNumber	 TargetNode,
	     TypeNumber	 EType);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Tests, whether there is a directed edge between	*/
		/*   the nodes.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern void
AGShowSourceNode(GraphNumber Graph,
		 NodeNumber  TargetNode,
		 TypeNumber  EType,
		 NodeNumber  *SourceNode);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns the source node to a given edge and target */
		/*   node.						*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   There must be one incoming edge of type "EType"	*/
		/*   at the node "TargetNode".				*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SIZE ERROR: there are too many open graphs		*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern void
AGTestAndShowSourceNode(GraphNumber Graph,
			NodeNumber  TargetNode,
			TypeNumber  EType,
			SHORTCARD   *SourceNrs,
			NodeNumber  *SourceNode);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   "SourceNrs" specifies, how many edges of type	*/
		/*   "EType" come into "TargetNode". If this count	*/
		/*   is exactly 1, the number of the source node is	*/
		/*   returned.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern void
AGShowAllSourceNodes(GraphNumber Graph,
		     NodeNumber	 TargetNode,
		     TypeNumber	 EType,
		     SimpleSet	 *SourceNodeSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns all numbers of source nodes to a given	*/
		/*   pair (edge,target node).				*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   The return set "SourceNodeSet" must have already	*/
		/*   been created.					*/
		/*							*/
		/*   Use "AGSetRemoveAnyElement" to access all nodes	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   UNKNOWN   : the set "SourceNodeSet" has not been	*/
		/*		 created				*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern void
AGShowTargetNode(GraphNumber Graph,
		 NodeNumber  SourceNode,
		 TypeNumber  EType,
		 NodeNumber  *TargetNode);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns the target node to a given edge and source */
		/*   node.						*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   There must be one outgoing edge of type "EType"	*/
		/*   at the node "SourceNode".				*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   SIZE ERROR: there is more than one target node	*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern void
AGShowAllTargetNodes(GraphNumber Graph,
		     NodeNumber	 SourceNode,
		     TypeNumber	 EType,
		     SimpleSet	 *TargetNodeSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns all numbers of target nodes to a given	*/
		/*   pair (source node,edge).				*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   The return set "TargetNodeSet" must have already	*/
		/*   been created.					*/
		/*							*/
		/*   Use "AGSetRemoveAnyElement" to access all nodes	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   UNKNOWN   : the set "TargetNodeSet" has not been	*/
		/*		 created				*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern void
AGTestAndShowTargetNode(GraphNumber Graph,
			NodeNumber  SourceNode,
			TypeNumber  EType,
			SHORTCARD   *TargetNrs,
			NodeNumber  *TargetNode);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   "TargetNrs" specifies, how many edges of type	*/
		/*   "EType" go out of "SourceNode". If this count	*/
		/*   is exactly 1, the number of the target node is	*/
		/*   returned.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern BOOLEAN
AGTestIncomingEdge(GraphNumber Graph,
		   NodeNumber  TargetNode,
		   TypeNumber  EType);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Tests, whether there is at least one edge of type	*/
		/*   "EType" coming into "TargetNode".		*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern BOOLEAN
AGTestOutgoingEdge(GraphNumber Graph,
		   NodeNumber  SourceNode,
		   TypeNumber  EType);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Tests, whether there is at least one edge of type	*/
		/*  "EType" going out of "SourceNode".			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   KIND ERROR: the edge type is not allowed		*/
		/*							*/
		/********************************************************/

extern void
AGGetInContextOfNode(GraphNumber Graph,
		     NodeNumber	 TargetNode,
		     RelSet	 *TupleSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   This procedure calculates a set of tuples of the	*/
		/*   form (EType,SourceNode). Each element of this	*/
		/*   set fulfills the condition, that the edge of type	*/
		/*   EType runs from SourceNode to the specified	*/
		/*   TargetNode.					*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   The return set "TupleSet" must have already been	*/
		/*   created.						*/
		/*							*/
		/*   Use "AGRelRemoveAnyTuple" to access all tuples	*/
		/*   stored within the temporarily existing "RelSet"	*/
		/*   and use "AGRelCreate" to initialize your relation	*/
		/*   before calling this procedure and use "AGRelKill"	*/
		/*   afterwards to release all storage occupied by this	*/
		/*   relation.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   UNKNOWN   : the relation "TupleSet" has not been	*/
		/*		 been created				*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*							*/
		/********************************************************/

extern void
AGGetOutContextOfNode(GraphNumber Graph,
		      NodeNumber  SourceNode,
		      RelSet	  *TupleSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   This procedure calculates a set of tuples of the	*/
		/*   form (EType,TargetNode). Each element of this	*/
		/*   set fulfills the condition, that the edge of type  */
		/*   EType runs from the specified SourceNode to	*/
		/*   TargetNode.					*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   The return set "TupleSet" must have already been	*/
		/*   created.						*/
		/*							*/
		/*   Use "AGRelRemoveAnyTuple" to access all tuples	*/
		/*   stored within the temporarily existing "RelSet"	*/
		/*   and use "AGRelCreate" to initialize your relation	*/
		/*   before calling this procedure and use "AGRelKill"	*/
		/*   afterwards to release all storage occupied by this	*/
		/*   relation.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   UNKNOWN   : the relation "TupleSet" has not been	*/
		/*		 been created				*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*							*/
		/********************************************************/

extern void
AGPutAttributeSubstr(GraphNumber     Graph,
		     NodeNumber	     Node,
		     AttributeNumber AttNo,
		     CARDINAL	     AttBegin,
		     CARDINAL	     AttLength,
		     const BYTE	     *AttValue,
		     BOOLEAN	     Truncate);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Replaces a substring of the AttNo'th (intrin-	*/
		/*   sic) attribute of a node by a new value. The sub-	*/
		/*   string begins at position "AttBegin" and has a 	*/
		/*   length of "AttLength" characters. The Truncate 	*/
		/*   argument specifies whether the attribute length 	*/
		/*   will be set to AttBegin+AttLength-1.		*/
		/*							*/
		/* Remarks:						*/
		/*							*/
		/* - A typical call of AGPutAttributeSubstr might look	*/
		/*   like "AGPutAttributeSubstr(g,n,0,0,SIZE(I),I,	*/
		/*   FALSE)", making use of SYSTEM's procedure SIZE.	*/
		/* - Writing a range starting at a position > 0 has the */
		/*   effect that reading below this position will not	*/
		/*   result in error messages or strings of length 0!	*/
		/*   The contents however may be undefined!		*/
		/* - The fact that AttValue is passed as a VAR para-	*/
		/*   meter has only efficiency reasons.			*/
		/* - The call "AGPutAttributeSubstr(g,n,X,0,0,I,TRUE)"	*/
		/*   will release the internal storage of the X'th	*/
		/*   attribute by setting its physical length to zero. 	*/
		/*   It's easier to use AGDeleteAttribute(g,n,X).	*/
		/* - index and key attributes must have a length up to	*/
		/*   250 bytes, normal attributes must be shorter 	*/
		/*   than 16449286 bytes (see constants MaxNormalAttri- */
		/*   buteLength and MaxIndexAttributeLength in module	*/
		/*   AGGlobal).						*/
		/* - The first about 250 bytes of the 0th normal attri-	*/
		/*   bute should be preferred by any application because*/
		/*   GRAS' internal storage mechanisms handle them in a */
		/*   more efficient way.				*/
                /*   If efficient storage of attribute values is essen- */
                /*   tial then please store all short attribute values  */
                /*   in UNDECLARED 0 attributes (the first attribute    */
                /*   value at position 0, the second at position 0 +    */
                /*   size of first attribute value etc.). But not that  */
                /*   maintaining indexes or incremental attribute eva-  */
                /*   luation doesn't work for these effficiently stored */
                /*   attributes.					*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*   NOT FOUND : the node "Node" does not exist		*/
		/*   SIZE ERROR: the value does not fit in the specified*/
		/*		 range					*/
		/*							*/
		/********************************************************/

extern void
AGDeleteAttribute(GraphNumber	  Graph,
		  NodeNumber	  Node,
		  AttributeNumber AttNo);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes the attribute AttNo from the node.		*/
		/*							*/
		/* Remarks:						*/
		/*							*/
		/*   A call to AGDeleteAttribute is basically a call	*/
		/*   AGPutAttributeSubstr(Graph,Node,AttNo,0,0,I,TRUE)	*/
		/*   (with a dummy parameter I - see AGPutAttribute-	*/
		/*   Substr).						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*   NOT FOUND : the node "Node" does not exist		*/
		/*   SIZE ERROR: the value does not fit in the specified*/
		/*		 range					*/
		/*							*/
		/********************************************************/

extern void
AGGetAttributeSubstr(GraphNumber     Graph,
		     NodeNumber	     Node,
		     AttributeNumber AttNo,
		     CARDINAL	     AttBegin,
		     CARDINAL	     AttLength,
		     BYTE	     *AttValue,
		     CARDINAL	     *ReturnedLength);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   This procedure performs the inverse action of	*/
		/*   AGPutAttributeSubstr. The value of the attribute	*/
		/*   substring is returned in the "generic" parameter	*/
		/*   "AttValue". ReturnedLength returns the number of	*/
                /*   bytes that could actually be read. If the value of */
                /*   the attribute is undefined (due to the fact that   */
                /*   the attribute didn't receive an initial value or   */
                /*   previous assignments - to a derived attribute -    */
                /*   became invalid due to certain graph changes), then */
                /*   ReturnedLength will be 0!                          */
		/*   In the case of an invalid derived attribute, the	*/
		/*   application may reevaluate this attribute and	*/
		/*   assign a new value (using 'AGPutAttributeS.')	*/
		/*   It is, however, possible to let GRAS do this	*/
		/*   automatically. In order to do this, information	*/
		/*   about the function name must be stored in the graph*/
		/*   scheme using "AGSetEvaluationFunction".		*/
		/*   Additionally, the application must specify the	*/
		/*   current address of the function code at runtime	*/
		/*   using "AGBindEvaluationFunction".			*/
		/*   In this case any attempt to read an invalid attri- */
		/*   bute will cause GRAS to reevaluate it and yield	*/
		/*   the new value.					*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   SIZE ERROR: the value does not fit in the specified*/
		/*		 range					*/
		/*   NOT FOUND : the node "Node" does not exist		*/
		/*   CYCLE     : cyclic attribute dependencies were	*/
		/*		 detected during automatic reevaluation */
		/*							*/
		/********************************************************/

/************************************************************************/
/*									*/
/*		      Operations for Event Handling			*/
/*									*/
/************************************************************************/

extern void
AGActivateDaemon(GraphNumber    Graph,
		 GraphEvent     Event,
		 ActionPriority Priority,
		 ActionProc     Action);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Activates a daemon. If "Event" occurs in the graph	*/
		/*   for which "Action" was defined, "Action" will be	*/
		/*   called with parameters as listed above. For every 	*/
		/*   tuple (Action,Event) there will be max. one daemon */
		/*   started. 						*/
		/*							*/
		/* Note:						*/
		/*  							*/
		/*  Let (k, n, t) be an occuring event. This event 	*/
		/*  matches an activated daemon if 			*/
		/*     Event.Kind = k					*/
		/*  /\ ( Event.TNo  = 0 \/ Event.TNo  = t )		*/
		/*  /\ ( Event.Node = 0 \/ Event.Node = n )		*/
		/*  i.e. 0 is a wildcard for TNo and Node fields of a 	*/
		/*  daemon.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   EXISTENT  : this (Action,Parameter,Event) triple	*/
		/*		 already exists				*/
		/*							*/
		/********************************************************/

extern void
AGKillDaemon(GraphNumber Graph,
	     GraphEvent	 Event,
	     ActionProc	 Action);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deactivates a daemon that has be activated by a	*/
		/*   call of AGActivateDaemon. If a graph is closed, all*/
		/*   its active daemons will be killed, so the next time*/
		/*   it is opened, there will be no initially active	*/
		/*   daemons.						*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   The execution of activated actions may not be	*/
		/*   locked for the specified graph!			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   NOT FOUND : the daemon does not exist		*/
		/*   UNKNOWN   : the action has not been defined	*/
		/*							*/
		/********************************************************/

extern void
AGRaiseEvent(GraphNumber Graph,
	     Surrogate p1, Surrogate p2, Surrogate p3);
                /********************************************************/
		/*                                                      */
		/* Purpose:                                             */
		/*  							*/
		/*  All daemons, that wait for UserDefined events	*/
		/*  (with matching parameters) will invoke their 	*/
		/*  actions. This can be used to communicate with	*/
		/*  other clients. But be careful: the meaning of	*/
		/*  p1 and p2 is totally undefined from GRAS and can	*/
		/*  be different from client to client.			*/
		/*                                                      */
                /* Note:                                                */
                /*                                                      */ 
                /*  RaiseEvent can also be called from a client         */ 
                /*  different from the owner of the graph who may have  */ 
                /*  defined the EventPattern and activated the daemon,  */
                /*  provided that he has the right to work on the graph */
                /*  under consideration.                                */
		/*                                                      */
                /********************************************************/

extern void
AGDelayActionExecution(GraphNumber Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Locks execution of all actions that were activated */
		/*   for the graph. Locking may be nested, which means	*/
		/*   that a certain number of AGDelayActionExecution	*/
		/*   calls must be followed be the same number of calls	*/
		/*   to AGReleaseActionExecution to release the lock.	*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   FORBIDDEN : execution is already locked		*/
		/*							*/
		/********************************************************/

extern void
AGReleaseActionExecution(GraphNumber Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Unlocks execution of activated actions for a graph.*/
		/*   Locked active actions are executed. Locking may be */
		/*   nested, which means that a certain number of 	*/
		/*   AGDelayActionExecution calls must be followed be 	*/
		/*   the same number of calls to AGReleaseAction-	*/
		/*   Execution to release the lock.			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   FORBIDDEN : execution is not locked		*/
		/*							*/
		/********************************************************/

/************************************************************************/
/*									*/
/*			Operations on Local Sets			*/
/*									*/
/* Note:								*/
/*									*/
/*   Aborting transactions does not effect local sets! They are not	*/
/*   reset into a previous state.					*/
/*									*/
/************************************************************************/

extern void
AGSetCreate(GraphNumber Graph,
	    SimpleSet   *EmptySet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Creates an empty temporary set which is local	*/
		/*   relative to the graph. A set is a structure with	*/
		/*   the properties of mathematical sets. It can store	*/
		/*   elements which are present at most once. Several	*/
		/*   operations are presented for these sets. SimpleSet	*/
		/*   structures contain SHORTCARD numbers.		*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   OVERFLOW  : creation failed because of internal	*/
		/*		 memory overflow			*/
		/*							*/
		/********************************************************/

extern void
AGSetUndefinedSet(GraphNumber Graph,
                  SimpleSet *set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*   							*/
		/*   Set becomes an undefined set. Never use undefined  */
                /*   sets in any set operations.			*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*							*/
		/********************************************************/

extern BOOLEAN      
AGSetIsUndefined(GraphNumber Graph,
                 SimpleSet set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*   							*/
		/*   Test if set is undefined   			*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*							*/
		/********************************************************/
      

extern void
AGSetClear(GraphNumber Graph,
	   SimpleSet   *ClearSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Clears an temporary set which is local relative to	*/
		/*   the graph. A set is a structure with the properties*/
		/*   of mathematical sets. It can store	elements which  */
		/*   are present at most once. Several operations are	*/
		/*   presented for these sets. SimpleSet structures con-*/
		/*   tain SHORTCARD numbers.				*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*							*/
		/********************************************************/

extern void
AGSetKill(GraphNumber Graph,
	  SimpleSet   *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Releases the amount of memory used by "Set". It	*/
		/*   may then be re-used be AGSetCreate.		*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGSetCopy(GraphNumber Graph,
	  SimpleSet   SourceSet,
	  SimpleSet   *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Copies "SourceSet" to "TargetSet", which must have	*/
		/*   been created before.				*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGSetInsertElement(GraphNumber Graph,
		   Surrogate   Element,
		   SimpleSet   *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Inserts an element into a set.			*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGSetDeleteElement(GraphNumber Graph,
		   Surrogate   Element,
		   SimpleSet   *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes an element from a set.			*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGSetRemoveAnyElement(GraphNumber Graph,
		      BOOLEAN	  *Found,
		      Surrogate	  *Element,
		      SimpleSet	  *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes an arbitrary element from the set and	*/
		/*   returns it to the calling module. If the set is	*/
		/*   empty, the result of "Found" is FALSE.		*/
		/*   "Arbitrary" means that the resulting sequence of   */
		/*   returned elements is the most efficient one with   */
		/*   respect to the system's internal storage layout.   */
		/*   Use "AGSetRemoveRndElement" if real random results */
		/*   are required.					*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGSetRemoveRndElement(GraphNumber Graph,
		      BOOLEAN	  *Found,
		      Surrogate	  *Element,
		      SimpleSet	  *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes a real random element from the set and	*/
		/*   returns it to the calling module. If the set is	*/
		/*   empty, the result of "Found" is FALSE.		*/
		/*   This procedure is considerably less efficient than */
		/*   "AGSetRemoveAnyElement" and should only be used in */
		/*   cases where "random" access is really important.   */
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern BOOLEAN
AGSetIsElement(GraphNumber Graph,
	       Surrogate   Element,
	       SimpleSet   Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns whether "Element" is in "Set".		*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern CARDINAL
AGSetSize(GraphNumber Graph,
	  SimpleSet   Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Computes the cardinal number of "Set".		*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGSetUnion(GraphNumber Graph,
	   SimpleSet   SourceSet,
	   SimpleSet   *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Computes the union of TargetSet and SourceSet and	*/
		/*   places the result in TargetSet.			*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGSetIntersection(GraphNumber Graph,
		  SimpleSet   SourceSet,
		  SimpleSet   *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Computes TargetSet := TargetSet intersected with	*/
		/*			   SourceSet			*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGSetDifference(GraphNumber Graph,
		SimpleSet   SourceSet,
		SimpleSet   *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Computes TargetSet := TargetSet without SourceSet. */
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

/************************************************************************/
/*									*/
/*		     Operations on Local Relations			*/
/*									*/
/* Note:								*/
/*									*/
/*   Aborting transactions does not effect local relations! They are	*/
/*   not reset into a previous state.					*/
/*									*/
/************************************************************************/

extern void
AGRelCreate(GraphNumber Graph,
	    RelSet      *EmptySet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Creates an empty temporary tuple set which is	*/
		/*   local relative to the graph. A relation set is a 	*/
		/*   structure with the properties of mathematical sets.*/
		/*   It can store elements which are present at most 	*/
		/*   once. Several operations are presented for these 	*/
		/*   sets. RelSet structures contain elements which are	*/
		/*   each a pair of SHORTCARD numbers.			*/
		/*							*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   OVERFLOW  : creation failed because of internal	*/
		/*		 memory overflow			*/
		/*							*/
		/********************************************************/

extern void
AGRelClear(GraphNumber Graph,
           RelSet      *ClearSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Clears an temporary tuple set which is		*/
		/*   local relative to the graph. A relation set is a 	*/
		/*   structure with the properties of mathematical sets.*/
		/*   It can store elements which are present at most 	*/
		/*   once. Several operations are presented for these 	*/
		/*   sets. RelSet structures contain elements which are	*/
		/*   each a pair of SHORTCARD numbers.			*/
		/*							*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*							*/
		/********************************************************/


extern void
AGRelKill(GraphNumber Graph,
	  RelSet      *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Releases the amount of memory used by "Set". It	*/
		/*   may then be re-used be AGRelCreate.		*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   UNKNOWN   : the set has not been created		*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*							*/
		/********************************************************/

extern void
AGRelCopy(GraphNumber Graph,
	  RelSet      SourceSet,
	  RelSet      *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Copies "SourceSet" to "TargetSet", which must have	*/
		/*   been created before.				*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGRelInsertTuple(GraphNumber Graph,
		 Surrogate   Surr1,
		 Surrogate   Surr2,
		 RelSet	     *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Inserts the Tuple ("Surr1","Surr2") into a set.	*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGRelDeleteTuple(GraphNumber Graph,
		 Surrogate   Surr1,
		 Surrogate   Surr2,
		 RelSet	     *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes the tuple ("Surr1","Surr2") from the set.	*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGRelRemoveAnyTuple(GraphNumber Graph,
		    BOOLEAN     *Found,
		    Surrogate   *Surr1,
		    Surrogate   *Surr2,
		    RelSet      *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes an arbitrary tuple from the set and	*/
		/*   returns its components "Surr1" and "Surr2" to the	*/
		/*   calling module. If the set is empty, the result of */
		/*   "Found" is FALSE.					*/
		/*   "Arbitrary" means that the resulting sequence of   */
		/*   returned elements is the most efficient one with   */
		/*   respect to the system's internal storage layout.   */
		/*   Use "AGRelRemoveRndElement" if real random results */
		/*   are required.					*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGRelRemoveRndTuple(GraphNumber Graph,
		    BOOLEAN     *Found,
		    Surrogate   *Surr1,
		    Surrogate   *Surr2,
		    RelSet      *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Deletes an arbitrary tuple from the set and	*/
		/*   returns its components "Surr1" and "Surr2" to the  */
		/*   calling module. If the set is empty, the result of */
		/*   "Found" is FALSE.					*/
		/*   This procedure is considerably less efficient than */
		/*   "AGRelRemoveAnyElement" and should only be used in */
		/*   cases where "random" access is really important.   */
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern BOOLEAN
AGRelIsElement(GraphNumber Graph,
	       Surrogate   Surr1,
	       Surrogate   Surr2,
	       RelSet	   Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   The result of this function is true, if the pair	*/
		/*   ("Surr1","Surr2") is in "Set".			*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern CARDINAL
AGRelSize(GraphNumber Graph,
	  RelSet      Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Computes the cardinal number of "Set".		*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGRelUnion(GraphNumber Graph,
	   RelSet      SourceSet,
	   RelSet      *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Computes the union of TargetSet and SourceSet and	*/
		/*   places the result in TargetSet.			*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGRelIntersection(GraphNumber Graph,
		  RelSet      SourceSet,
		  RelSet      *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Computes TargetSet := TargetSet intersected with	*/
		/*			   SourceSet			*/
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGRelDifference(GraphNumber Graph,
		RelSet	    SourceSet,
		RelSet	    *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Computes TargetSet := TargetSet without SourceSet. */
		/*							*/
		/* Errors Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGCartesian(GraphNumber Graph,
	    SimpleSet   SourceSet1,
	    SimpleSet   SourceSet2,
	    RelSet      *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Computes the cartesian product of the two source	*/
		/*   sets and places the result in the tuple set	*/
		/*   "TargetSet", which must have been created before.	*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGProjectionFirst(GraphNumber Graph,
		  RelSet      SourceSet,
		  SimpleSet   *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Projection of the first component of "SourceSet"	*/
		/*   into "TargetSet", which must have been created 	*/
		/*   before.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGProjectionSecond(GraphNumber Graph,
		   RelSet      SourceSet,
		   SimpleSet   *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Projection of the second component of "SourceSet"	*/
		/*   into "TargetSet", which must have been created 	*/
		/*   before.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGQueryProjection(GraphNumber Graph,
		  RelSet      SourceSet,
		  Surrogate   Surr,
		  SimpleSet   *TargetSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Projection of the second component of "SourceSet"	*/
		/*   into "TargetSet", if the first component matches	*/
		/*   "Surr". "TargetSet" must have been created before. */
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : one of the sets has not been created	*/
		/*							*/
		/********************************************************/

extern void
AGSingleQueryProjection(GraphNumber Graph,
			RelSet	    Set,
			Surrogate   Surr1,
			Surrogate   *Surr2,
			BOOLEAN	    *Found);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   AGSingleQueryProjection calculates *one* second	*/
		/*   component Surr2 for a given first component Surr1. */
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGChange(GraphNumber Graph,
	 RelSet	     *Set);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Exchanges the components of a local tuple set.	*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

/************************************************************************/
/*									*/
/*		      Procedures to work with Logs			*/
/*									*/
/************************************************************************/

extern void
AGSetCheckpoint(GraphNumber Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Sets a checkpoint.					*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*   PENDING T.: there is an active transaction running */
		/*		 on the graph "Graph"			*/
		/*							*/
		/********************************************************/

extern void
AGUndo(GraphNumber Graph,
       CARDINAL	   *N);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   AGUndo performs an UNDO back for N check points.	*/
		/*   If there are less than N check points set, the	*/
		/*   number of successful UNDOs is returned in N. This	*/
		/*   procedure should only be called, if there were no	*/
		/*   graph changing operations since the last setting	*/
		/*   of a check point. Otherwise, a new checkpoint is	*/
		/*   created internally before executing AGUndo itself.	*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*   PENDING T.: there is an active transaction running */
		/*		 on the graph "Graph"			*/
		/*							*/
		/********************************************************/

extern void
AGRedo(GraphNumber Graph,
       CARDINAL	   *N);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   AGRedo performs a REDO of all graph changing	*/
		/*   operations that were performed since N check	*/
		/*   points before. This procedure may only be called,	*/
		/*   if there were no graph changing operations since	*/
		/*   the last setting of a check point. The number of	*/
		/*   successful REDOs is returned as parameter N.	*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*   FORBIDDEN : the last command was a write command	*/
		/*   PENDING T.: there is an active transaction running */
		/*		 on the graph "Graph"			*/
		/*							*/
		/********************************************************/

/************************************************************************/
/*									*/
/*			     Scheme Management				*/
/*									*/
/*  N O T E :  The following procedures are not necessary for the mani- */
/*	       pulation of attributed graphs (if the error check mode	*/
/*	       is deactivated and incremental attribute evaluation as	*/
/*	       well as index or key attributes are not used).		*/
/*									*/
/************************************************************************/

/************************************************************************/
/*									*/
/*	   Operations for managing schemes				*/
/*									*/
/* Every component of the graph scheme has two identifiers. One is a	*/
/* (human-readable) name, the other one an arbitrary internally used	*/
/* number which must be unique across all types of declared graph	*/
/* compontents.								*/
/*                                                                      */
/* ATTENTION: Even components belonging to different categories of data */
/*            - like node classes and edge types - must have different  */
/*            names and internally used type numbers.                   */
/*									*/
/* This part of the GRAS interface is not yet complete. Explanations of */
/* error messages must be added and a number of procedures for querying	*/
/* an already existing graph are missing.				*/
/*									*/
/************************************************************************/

extern void
AGDeclareScheme(GraphNumber  Graph,
		const CHAR   *Name,
		SchemeNumber *Scheme);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Creation of a scheme with the name "Name" that	*/
		/*   will get the number "Scheme". This number can be	*/
		/*   used to reference it. It is possible to have more	*/
		/*   than one (sub-)scheme in one graph.		*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   AGDeclareScheme may be used to extend already      */
                /*   existing graph schemes and to redefine the         */
                /*   evaluation functions and attribute dependencies    */
                /*   of previously declared attributes, i.e. AGDeclare- */
		/*   Scheme for an already existing graph scheme starts */
		/*   the extension of this graph scheme.		*/
		/*							*/
		/********************************************************/

extern void
AGCommitScheme(GraphNumber  Graph,
	       SchemeNumber Scheme);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Commit all declarations that were made for this	*/
		/*   scheme. This procedure is used to tell the GRAS	*/
		/*   system that there are no more declarations. There- */
		/*   fore it must be called AFTER all other scheme	*/
		/*   procedures. All declarations are then carried out	*/
		/*   in the proper order. That means, that you can call	*/
		/*   all declarations in arbitrary order and then simply*/
		/*   commit the scheme.					*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   AGCommitScheme may be used to extend already       */
                /*   existing graph schemes and to redefine the         */
                /*   evaluation functions and attribute dependencies    */
                /*   of previously declared attributes.  		*/
		/*							*/
		/********************************************************/

extern void
AGTestAndShowSchemeNumber(GraphNumber  Graph,
			  const CHAR   *Name,
			  BOOLEAN      *Existent,
			  SchemeNumber *Scheme);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Look for the scheme "Name". If it is existent, its */
		/*   number is returned in "Scheme".			*/
		/*							*/
		/********************************************************/

extern void
AGShowSchemeName(GraphNumber  Graph,
		 SchemeNumber Scheme,
		 CHAR	      *Name);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get the name of the scheme "Scheme".		*/
		/*							*/
		/********************************************************/

extern void
AGSetSchemeVersion(GraphNumber  Graph,
		   SchemeNumber Scheme,
		   const CHAR   *Version);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Set the version of the scheme "Scheme".		*/
		/*							*/
		/********************************************************/

extern void
AGShowSchemeVersion(GraphNumber	 Graph,
		    SchemeNumber Scheme,
		    CHAR	 *Version);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get the version of the scheme "Scheme".		*/
		/*							*/
		/********************************************************/

/*****************************************************************************/
/*									     */
/*  Procedures for managing node classes				     */
/*									     */
/*****************************************************************************/

extern void
AGDeclareNodeClass(GraphNumber  Graph,
		   SchemeNumber Scheme,
		   const CHAR   *Name,
		   TypeNumber   NodeClass);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Declare the node class "Name" for the scheme	*/
		/*   "Scheme". This node class will get the number	*/
		/*   "NodeClass". The name and the number of the node	*/
		/*   class must be unique for this graph.		*/
		/*   Node class no. 0 is reserved.			*/
		/*							*/
		/* Note:						*/
		/*   This action will not take place before the scheme	*/
		/*   is committed.					*/
		/*							*/
		/********************************************************/

extern void
AGTestAndShowNodeClassNumber(GraphNumber Graph,
			     const CHAR  *Name,
			     BOOLEAN	 *Existent,
			     TypeNumber	 *NodeClass);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Look for the node class "Name". If it is existent, */
		/*   its number is returned in "NodeClass".		*/
		/*							*/
		/********************************************************/

extern void
AGShowNodeClassName(GraphNumber Graph,
		    TypeNumber  NodeClass,
		    CHAR        *Name,
		    BOOLEAN     *Existent);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get the name of the node class "NodeClass". If the */
		/*   node class is declared, the name is returned in	*/
		/*   "Name".						*/
		/*							*/
		/********************************************************/

extern void
AGShowAllNodeTypes(GraphNumber Graph,
		   TypeNumber  NodeClass,
		   SimpleSet   *SetOfTypes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*  Get all node types of a given class and all of its  */
		/*  subclasses in "SetOfTypes" (defined but empty set), */
		/*  i.e. the set must have been created before.		*/
		/*							*/
		/*   Note:						*/
		/*   							*/
		/*   Use "AGSetRemoveAnyElement" to access all types	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/********************************************************/

extern void
AGAppendNodeClass(GraphNumber  Graph,
		  SchemeNumber Scheme,
		  const CHAR   *NodeClass,
		  const CHAR   *SuperClass);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Add a specialization/generalization relation to	*/
		/*   the scheme. The node class "NodeClass" will then	*/
		/*   be a subclass of "SuperClass" and inherits all	*/
		/*   attributes and attribute evaluation rules.		*/
		/*							*/
		/* Note:						*/
		/*   This action will not take place before the scheme	*/
		/*   is committed.					*/
		/*							*/
		/********************************************************/

extern void
AGShowDirectSuperclasses (GraphNumber Graph,
			  TypeNumber NodeClassOrType,
			  SimpleSet  *SetOfClasses    );
		/********************************************************/
		/*                                                      */
		/* Purpose:                                             */
		/*                                                      */
		/*  Get all direct super classes of a given class or    */
		/*  type in "SetOfClasses" (defined but empty set),     */
		/*  i.e. the set must have been created before.         */
		/*                                                      */
		/*   Note:                                              */
		/*                                                      */
		/*   Use "AGSetRemoveAnyElement" to access all types    */
		/*   stored within the temporarily existing "SimpleSet" */
		/*   and use "AGSetCreate" to initialize the set before */
		/*   calling this procedure and use "AGSetKill" after-  */
		/*   wards to release all storage occupied by this set. */
		/*                                                      */
		/********************************************************/

extern void
AGShowDirectSubclasses (GraphNumber Graph,
			TypeNumber NodeClass,
			SimpleSet *SetOfClasses    );
		/********************************************************/
		/*                                                      */
		/* Purpose:                                             */
		/*                                                      */
		/*  Get all direct sub classes  of a given              */
		/*  clas in "SetOfClasses" (defined but empty           */
		/*  set),i.e. the set must have been created before.    */
		/*                                                      */
		/*   Note:                                              */
		/*                                                      */
		/*   Use "AGSetRemoveAnyElement" to access all types    */
		/*   stored within the temporarily existing "SimpleSet" */
		/*   and use "AGSetCreate" to initialize the set before */
		/*   calling this procedure and use "AGSetKill" after-  */
		/*   wards to release all storage occupied by this set. */
		/*                                                      */
		/********************************************************/

extern void
AGShowDirectSubtypes (GraphNumber Graph,
		      TypeNumber NodeClass,
		      SimpleSet *SetOfTypes    );
		/********************************************************/
		/*                                                      */
		/* Purpose:                                             */
		/*                                                      */
		/*  Get all direct sub types of a given                 */
		/*  clas in "SetOfClasses" (defined but empty           */
		/*  set),i.e. the set must have been created before.    */
		/*                                                      */
		/*   Note:                                              */
		/*                                                      */
		/*   Use "AGSetRemoveAnyElement" to access all types    */
		/*   stored within the temporarily existing "SimpleSet" */
		/*   and use "AGSetCreate" to initialize the set before */
		/*   calling this procedure and use "AGSetKill" after-  */
		/*   wards to release all storage occupied by this set. */
		/*                                                      */
		/********************************************************/

/*****************************************************************************/
/*									     */
/* Procedures for managing node types					     */
/*									     */
/*****************************************************************************/

extern void
AGDeclareNodeType(GraphNumber  Graph,
		  SchemeNumber Scheme,
		  BOOLEAN      UseMembership,
		  const CHAR   *Name,
		  TypeNumber   NType);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Declare node type "Name" for scheme "Scheme". This */
		/*   node type will get  number "NType". The name 	*/
		/*   and the number of the node type must be unique for	*/
		/*   this graph. Node type no. 0 is reserved.		*/
		/*   If the "UseMembership" flag is set, GRAS will be	*/
		/*   able to find all nodes of that type using the	*/
		/*   procedure "AGGetAllNodesOfNodeType".		*/
		/*   The "UseMembership" flag may not be used if more	*/
		/*   than 5000 nodes of this type will be created.	*/
		/*							*/
		/* Notes:						*/
		/*   1) This action will not take place before the	*/
		/*	scheme is committed.				*/
		/*   2) Creating and deleting nodes of this type	*/
		/*	will be more efficient if the		 	*/
		/*      "UseMembership" option is NOT set.		*/
		/*							*/
		/********************************************************/

extern void
AGTestAndShowNodeTypeNumber(GraphNumber Graph,
			    const CHAR  *Name,
			    BOOLEAN     *Existent,
			    TypeNumber  *NType);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Look for the node type "Name". If it is existent,	*/
		/*   its number is returned in "NType".  		*/
		/*							*/
		/********************************************************/

extern void
AGShowNodeTypeName(GraphNumber Graph,
		   TypeNumber  NType,
		   CHAR	       *Name,
		   BOOLEAN     *Existent);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get the name of the node type "NType". If the	*/
		/*   node type was declared, the name is returned in	*/
		/*   "Name".						*/
		/*							*/
		/********************************************************/

extern void
AGAppendNodeType(GraphNumber  Graph,
		 SchemeNumber Scheme,
		 const CHAR   *NType,
		 const CHAR   *NodeClass);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Add a specialization/generalization relation to	*/
		/*   the scheme. The node type "NType" will then	*/
		/*   belong to the class "NodeClass" and inherits all	*/
		/*   attributes and attribute evaluation rules.		*/
		/*							*/
		/* Note:						*/
		/*   This action will not take place before the scheme	*/
		/*   is committed.					*/
		/*							*/
		/********************************************************/

/*****************************************************************************/
/*									     */
/* Procedures for managing edge types					     */
/*									     */
/*****************************************************************************/

extern void
AGDeclareEdgeType(GraphNumber  Graph,
		  SchemeNumber Scheme,
		  const CHAR   *Name,
		  const CHAR   *SourceNodeClassOrType,
		  Cardinality  SourceCardinality,
		  const CHAR   *TargetNodeClassOrType,
		  Cardinality  TargetCardinality,
	          TypeNumber   Edge);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Declare the edge type "Name" for 			*/
		/*   the scheme "Scheme". The edge type will get the 	*/
		/*   number "Edge". Edge type no. 0 is reserved.	*/
		/*   All nodes that belong to "SourceNodeClassOrType"	*/
		/*   are allowed as source nodes for edges of this type.*/
		/*   All nodes that belong to "TargetNodeClassOrType"	*/
		/*   are allowed as target nodes for edges of this type.*/
		/*   "SourceCardinality" and "TargetCardinality" are    */
		/*   used for runtime checks. "TargetCardinality" =     */
		/*   "ObligateSet" means e.g. that any node of class    */
		/*   "SourceNodeClassOrType" is expected to have at     */
		/*   least one emanating edge of this type.             */ 
		/*							*/
		/* Note:						*/
		/*   This action will not take place before the scheme	*/
		/*   is committed.					*/
		/* 							*/
		/* ATTENTION:						*/
		/*   Up to now, cardinality checks in "AGCreateEdge"    */
		/*   etc. are not yet implemented.			*/
		/*							*/
		/********************************************************/

extern void
AGTestAndShowEdgeTypeNumber(GraphNumber Graph,
			    const CHAR  *Name,
			    BOOLEAN     *Existent,
			    TypeNumber  *Edge);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Look for the edge type "Name". If it is existent,	*/
		/*   its number is returned in "Edge".			*/
		/*							*/
		/********************************************************/

extern void
AGShowEdgeTypeName(GraphNumber Graph,
		   TypeNumber  Edge,
		   CHAR	       *Name,
		   BOOLEAN     *Existent);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get the name of the edge type "Edge". If the	*/
		/*   edge type was declared, the name is returned in	*/
		/*   "Name".						*/
		/*							*/
		/********************************************************/

extern void
AGShowEdgeTypeProps(GraphNumber Graph,
		    TypeNumber  Edge,
		    TypeNumber  *SourceCT,
		    Cardinality *SourceCardinality,
		    TypeNumber  *TargetCT,
		    Cardinality *TargetCardinality,
		    BOOLEAN     *Existent);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get the properties of the edge type "Edge". If the	*/
		/*   edge type was declared, its cardinalities and its  */
		/*   source and target types/classes are returned.	*/
		/*							*/
		/********************************************************/

/*****************************************************************************/
/*									     */
/* Procedures for managing attributes					     */
/*									     */
/*****************************************************************************/

extern void
AGDeclareAttribute(GraphNumber	   Graph,
		   SchemeNumber	   Scheme,
		   const CHAR	   *ClassOrType,
		   const CHAR	   *Name,
		   AttributeKind   Kind,
		   IndexProperties Props,
		   ValueTypeNumber ValueType,
		   Cardinality	   ValueCardinality,
		   BOOLEAN	   ConstantLength,
		   CARDINAL	   Length,
		   AttributeNumber AttNo);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Declare the Attribute "Name" for the scheme	*/
		/*   "Scheme". All nodes that belong to the class/type	*/
		/*   "ClassOrType" will carry that attribute.           */
		/*   The attribute will get the number "AttNo". The	*/
		/*   name and the number of the attribute must be	*/
		/*   unique.						*/
		/*							*/
		/*   "Kind" specifies whether the attribute is derived	*/
		/*   (with its subcase dummy), i.e. depends on other    */
		/*   attributes, or whether it is an intrinsic attribute*/
		/*   (with its subcase meta attribute which has the same*/
		/*   value for all nodes of a certain type, i.e. is an  */
		/*   attribute of the node type itself).		*/
		/*							*/
		/*   If constant length flag is set, "Length" specifies	*/
		/*   the size of the values that will be stored in 	*/
		/*   that attribute (in Bytes).				*/
		/*							*/
		/*   "Props" contains information about the properties	*/
		/*   of the attribute, i.e. whether it is a normal, an	*/
		/*   index or a key attribute.				*/
		/*							*/
		/*   "ValueType" is an almost uninterpreted number which*/
		/*   gives applications the chance to store information */
		/*   about the type of an attribute's value (which is   */
		/*   normally considered to be an uninterpreted byte    */
		/*   stream. The numbers for standard types like integer*/
		/*   boolean etc. are fixed! All remaining numbers may  */
		/*   be used as "pointers" to application defined types.*/
		/*							*/
		/*   "ValueCardinality" specifies whether a value for	*/
		/*   this attribute is optional ( -> "Opt..." ) or	*/
		/*   obligate ( -> "Obl..." ) and whether the attribute */
		/*   contains a set of values ( -> "...Set" ) or only 	*/
		/*   a single element ( -> "...Unique" )		*/
		/*   Note that "ValueCardinality" is an attribute prop.	*/
		/*   which has no meaning for GRAS (like "ValueType")	*/
		/*   because (up to now) all (sets of) attribute values	*/
		/*   are uninterpreted arrays of byte for GRAS.		*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   1) The attribute number 0 should be used for the   */
		/*      most frequently used NON-INDEX (and NON-KEY)    */
		/*      attribute. If time and space consumption are    */
		/*      very important, then you might even store many  */
		/*      short attribute values within the 0 attribute   */
		/*      (or within another attribute) without declaring */
		/*  	this attribute. This trick deactivates a number */
		/*	of checks (for keeping derived attributes and   */
		/*	attribute indexes consistent).			*/
		/*   2) Meta and dummy index or key attributes are not	*/
		/*      supported.					*/
		/*   3) Dummy attributes may be used for propagating	*/
		/*	graph changes along arbitrarily long paths in	*/
		/*	a graph (by translating an n-context dependency	*/
		/*      between two attributes into n 1-context depen-  */
		/*	dencies between dummy attributes).		*/
		/*   4) Dummy attributes only are allowed to have a     */
		/*      constant length of size 0.			*/
		/*							*/
		/********************************************************/

extern void
AGTestAndShowAttributeNumber(GraphNumber     Graph,
			     const CHAR      *Name,
			     BOOLEAN	     *Existent,
			     AttributeNumber *AttNo);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Look for the attribute "Name". If it is existent,	*/
		/*   its number is returned in "AttNo".			*/
		/*							*/
		/********************************************************/

extern void
AGShowAttributeName(GraphNumber	    Graph,
		    AttributeNumber AttNo,
		    CHAR	    *Name,
		    BOOLEAN	    *Existent);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get the name of the attribute "AttNo". If the	*/
		/*   attribute was declared, the name is returned in	*/
		/*   "Name".						*/
		/*							*/
		/********************************************************/

extern void
AGShowAttributeProps(GraphNumber     Graph,
		     AttributeNumber AttNo,
		     AttributeKind   *Kind,
		     IndexProperties *Props,
		     ValueTypeNumber *ValueType,
		     Cardinality     *AttCard,
		     BOOLEAN	     *ConstLength,
		     CARDINAL	     *Length,
		     BOOLEAN	     *Existent);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get the properties of the attribute "AttNo".	*/
		/*							*/
		/********************************************************/

extern void
AGPutMetaAttribute(GraphNumber	   Graph,
		   TypeNumber	   NodeType,
		   AttributeNumber AttNo,
		   CARDINAL	   AttBegin,
		   CARDINAL	   AttLength,
		   const BYTE	   *AttValue,
		   BOOLEAN	   Truncate);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Basically works like 'AGPutAttributeSubstr' but is */
		/*   used to define the "meta" attribute value of	*/
		/*   'AttNo' at a node type object (and not at a node). */
		/*							*/
		/********************************************************/

extern void
AGGetMetaAttribute(GraphNumber	   Graph,
		   TypeNumber	   NodeType,
		   AttributeNumber AttNo,
		   CARDINAL	   AttBegin,
		   CARDINAL	   AttLength,
		   BYTE		   *AttValue,
		   CARDINAL	   *ReturnedLength);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Basically works like 'AGGetAttributeSubstr' but is */
		/*   used to access the "meta" attribute value of	*/
		/*   'AttNo' at a node type object (and not at a node)  */
		/*   and is therefore the inverse function to "AGPut-   */
		/*   MetaAttribute".					*/
		/*							*/
		/********************************************************/

extern void
AGSetEvaluationFunction(GraphNumber  Graph,
			SchemeNumber Scheme,
			const CHAR   *SetForCT,
			const CHAR   *SetForAttribute,
			const CHAR   *Name);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Set the name of the evaluation function for the	*/
		/*   attribute "Attribute". This name will be		*/
		/*   valid for all nodes that belong to the class/type	*/
		/*   "NodeClassOrType" with the exception of those 	*/
		/*   attributes which belong to a subclass (type) with	*/
		/*   a redefinition of the evaluation function.		*/
		/*   To get the valid evaluation function for a specific*/
		/*   class or type use "AGShowEvaluationFunction".	*/
		/*   The purpose of an evaluation function depends on	*/
		/*   the attribute's kind:				*/
		/*	derived:   Recompute an attribute's value.	*/
		/*	intrinsic: Initialize new attribute instances	*/
		/*		   (not for intrinsic KEY attribute).   */
		/*	meta:      Compute constant value for all nodes */
		/*		   of given node type (i.e. the 'Name'  */
		/*		   function has a node type instead of  */
		/* 		   a node as one of its parameters).    */
		/*	dummy:	   Not useful!				*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   1) This action will not take place before the	*/
		/*	scheme is committed.				*/
		/*   2) The name of the evaluation function may be	*/
		/*	redefined in subclasses/subtypes.		*/
		/*   3) You have to use "AGBindEvaluationFunction" to	*/
		/*	provide GRAS at runtime with the code addresses	*/
		/*	of these functions.				*/
		/*							*/
		/* Feature:						*/
		/*							*/
		/*   The main purpose of an evaluation function is to   */
		/*   compute a new value for an invalid attribute and   */
		/*   to store this value within the proper attribute    */
		/*   itself by using 'AGPutAttributeSubstr'.		*/
		/*							*/
		/********************************************************/

extern void
AGShowEvaluationFunction(GraphNumber	 Graph,
			 AttributeNumber AttNo,
			 TypeNumber	 NodeClassOrType,
			 CHAR		 *FunctionName,
			 BOOLEAN	 *Existent);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get the name of the evaluation function for the	*/
		/*   attribute "AttNo" and the node class/type		*/
		/*   "NodeClassOrType". If the evaluation		*/
		/*   function was set, the name is returned in		*/
		/*   "Name".						*/
		/*							*/
		/********************************************************/

extern void
AGBindEvaluationFunction(GraphNumber  Graph,
			 const CHAR   *Name,
			 EvalFunction Function);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Specify the current address of the evaluation	*/
		/*   function "Name" at runtime.			*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   This function must be used after each "AGLogin" to	*/
		/*   provide GRAS at runtime with the code addresses	*/
		/*   of the evaluation functions.			*/
		/*							*/
		/********************************************************/

extern void
AGDeclareDependency(GraphNumber	   Graph,
		    SchemeNumber   Scheme,
		    const CHAR     *NodeClassOrType,
		    const CHAR     *Dependent,
		    const CHAR     *DependsOn,
		    const CHAR     *NameOfEdge,
		    DependencyKind KindOfDependency);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Statically declare a dependency of the attribute	*/
		/*   "Dependent" from the attribute "DependsOn". This	*/
		/*   dependency will be valid for all nodes that belong */
		/*   to the class/type "NodeClassOrType".		*/
		/*   "KindOfDependency" specifies the kind of dependency*/
		/*   of the attribute (see AGGlobal for details).	*/
		/*   GRAS will delete all attributes that are affected	*/
		/*   by attribute writing or structural changes of the	*/
		/*   graph. This values must be reevaluated later on.	*/
		/*							*/
		/* Note:						*/
		/*   This action will not take place before the scheme	*/
		/*   is committed.					*/
		/*							*/
		/********************************************************/

extern void
AGDeleteDependencies(GraphNumber Graph,
		     const CHAR  *NodeClassOrType,
		     const CHAR  *Dependent,
		     BOOLEAN	 RestoreInheritance);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Delete all dependencies of the attribute		*/
		/*   "Dependent" for the node class/type		*/
		/*   "NodeClassOrType". This will also delete		*/
		/*   any evaluation functions that were declared for	*/
		/*   this attribute/class combination.			*/
		/*   If "RestoreInheritance" is set, this class will	*/
		/*   get the dependency status and evaluation function	*/
		/*   from its closest superclass(es).			*/
		/*							*/
		/********************************************************/

extern void
AGShowAllAttributesOfNodeType(GraphNumber Graph,
			      TypeNumber  NType,
			      SimpleSet	  *Attributes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get all attributes that were declared for the node */
		/*   type "NType". The set must have been created.	*/
		/*							*/
		/*   Note:						*/
		/*   							*/
		/*   Use "AGSetRemoveAnyElement" to access all attnos.	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/********************************************************/

extern void
AGGetAllNodesOfGraph(GraphNumber Graph,
		     SimpleSet	 *SetOfNodes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get all nodes in the graph. The set must have been */
		/*   created. All node types must use the membership	*/
		/*   relation.						*/
		/*							*/
		/*   Note:						*/
		/*   							*/
		/*   Use "AGSetRemoveAnyElement" to access all nodes	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/********************************************************/

extern void
AGGetAllNodeclasses(GraphNumber Graph,
		    SimpleSet *SetOfNodes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get all declared classes in the graph. The set     */
                /*   must have been created.                            */
		/*   							*/
		/*   Note:						*/
		/*   							*/
		/*   Use "AGSetRemoveAnyElement" to access all classes	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/********************************************************/


extern void
AGGetAllEdgetypes(GraphNumber Graph,
		  SimpleSet *SetOfNodes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get all declared edge types in the graph. The set  */
                /*   must have been created.                            */
		/*   							*/
		/*   Note:						*/
		/*   							*/
		/*   Use "AGSetRemoveAnyElement" to access all types    */
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/********************************************************/



extern void
AGGetAllNodesOfNodeClass(GraphNumber Graph,
			 TypeNumber  NodeClass,
			 SimpleSet   *SetOfNodes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get all nodes of the class "NodeClass". The set	*/
		/*   must have been created. All node types must use	*/
		/*   the membership relation.				*/
		/*							*/
		/*   Note:						*/
		/*   							*/
		/*   Use "AGSetRemoveAnyElement" to access all nodes	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/********************************************************/

extern void
AGGetAllNodesOfNodeType(GraphNumber Graph,
			TypeNumber  NType,
			SimpleSet   *SetOfNodes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get all nodes of type "NType". The set		*/
		/*   must have been created. All subtypes must use	*/
		/*   the membership relation.				*/
		/*							*/
		/*   Note:						*/
		/*   							*/
		/*   Use "AGSetRemoveAnyElement" to access all nodes	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/********************************************************/

extern void
AGGetComplementOfNodeSet(GraphNumber Graph,
			 SimpleSet   *SetOfNodes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Get all nodes of the graph that do not belong to	*/
		/*   "SetOfNodes". All node types must use the		*/
		/*   membership relation.				*/
		/*							*/
		/*   Note:						*/
		/*   							*/
		/*   Use "AGSetRemoveAnyElement" to access all nodes	*/
		/*   stored within the temporarily existing "SimpleSet"	*/
		/*   and use "AGSetCreate" to initialize the set before	*/
		/*   calling this procedure and use "AGSetKill" after-	*/
		/*   wards to release all storage occupied by this set. */
		/*							*/
		/********************************************************/

extern BOOLEAN
AGTestIsNodeOfNodeClass(GraphNumber Graph,
			TypeNumber  NodeClass,
			TypeNumber  Node);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Self explanatory.					*/
		/*							*/
		/********************************************************/

/*****************************************************************************/
/*									     */
/* Procedures for setting the error check mode				     */
/*									     */
/*****************************************************************************/

extern void
AGSetErrorCheckMode(GraphNumber Graph,
		    BOOLEAN     ErrorcheckOn);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Switch on/off the runtime legality tests. The	*/
		/*   default setting is OFF.				*/
		/*   This affects only expensive checks against graph	*/
		/*   schemes, i.e. even with error check "OFF"  many	*/
		/*   consistency checks are active in order to avoid	*/
		/*   "serious" corruptions of graphs. Set this switch	*/
		/*   to "ON" (TRUE) for debugging purposes and to "OFF"	*/
		/*   (FALSE) in your production version.		*/
		/*							*/
		/********************************************************/

  /**********************************************************************/
  /*                                                                    */
  /* Here starts the 'hack section'! The procedures below are to be	*/
  /* handled with care because they may lead to inconsistent states	*/
  /* (if not carefully used) which may result in loss of data!		*/
  /*                                                                    */
  /* The reasons for still offering these procedures are twofold:	*/
  /*	(1) Compatibility with older GRAS versions			*/
  /*    (2) Efficient usage of GRAS resources in the absence of a graph	*/
  /*	    scheme							*/
  /*                                                                    */
  /*                               *****                                */
  /*                               *   *                                */
  /*                               *   *                                */
  /*                               *   *                                */
  /*                               *   *                                */
  /*                            ****   ****                             */
  /*                             *       *                              */
  /*                              *     *                               */
  /*                               *   *                                */
  /*                                * *                                 */
  /*                                 *                                  */
  /*                                                                    */
  /**********************************************************************/

extern void
AGPutIndexAttribute(GraphNumber	    Graph,
		    NodeNumber	    Node,
		    AttributeNumber AttNo,
		    CARDINAL	    AttLength,
		    const BYTE	    *AttValue);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Without scheme					*/
		/*   --------------					*/
		/*   Assigns a new value to an index attribute and over-*/
		/*   rides the old value, if existent.			*/
		/*							*/
		/*   With scheme					*/
		/*   -----------					*/
		/*   Creates an index attribute for "Node". If the	*/
		/*   scheme is used, we recommend to use AGPutAttribute-*/
		/*   Substr rather than this procedure!			*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   AttValue is a VAR parameter for efficiency reasons */
		/*   only.						*/
		/*   Don't use AttNo=0! (reserved for non-index att.)	*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   NOT FOUND : the node "Node" does not exist		*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*							*/
		/********************************************************/

extern void
AGGetIndexAttribute(GraphNumber	    Graph,
		    NodeNumber	    Node,
		    AttributeNumber AttNo,
		    BYTE	    *AttValue,
		    CARDINAL	    *AttLength);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Without scheme					*/
		/*   --------------					*/
		/*   Returns the value of the attribute "AttNo".	*/
		/*							*/
		/*   With scheme					*/
		/*   -----------					*/
		/*   Returns the value of an index attribute for Node.	*/
		/*   If the scheme is used, we recommend to use AGGet-	*/
		/*   AttributeSubstr rather than this procedure!	*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   NOT FOUND : the node "Node" does not exist		*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*							*/
		/********************************************************/

extern void
AGDeleteIndexAttribute(GraphNumber     Graph,
		       NodeNumber      Node,
		       AttributeNumber AttNo);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Without scheme					*/
		/*   --------------					*/
		/*   Deletes an index attribute from a node.		*/
		/*							*/
		/*   With scheme					*/
		/*   -----------					*/
		/*   Deletes an index attribute from a node. If the 	*/
		/*   scheme is used, we recommend to use AGPutAttribute-*/
		/*   Substr rather than this procedure!			*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   NOT FOUND : the node does not exist or an attribute*/
		/*		 with a number "AttNo" has no defined	*/
		/*		 value					*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   LOCKED    : the graph has been opened for read	*/
		/*		 access only				*/
		/*							*/
		/********************************************************/

extern void
AGGetAllIndexAttributes(GraphNumber Graph,
			NodeNumber  Node,
			SimpleSet   *IndexSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Without scheme					*/
		/*   --------------					*/
		/*   Returns all attribute numbers of index attributes	*/
		/*   corresponding to node whose value is currently	*/
		/*   defined.						*/
		/*							*/
		/*   With scheme					*/
		/*   -----------					*/
		/*   Returns all index attribute numbers corresponding 	*/
		/*   to Node which are NOT declared in the graph scheme.*/
		/*							*/
		/*   In both cases the set must have been created	*/
		/*   before.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   NOT FOUND : the node "Node" does not exist		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGGetAllDefinedAttributes(GraphNumber Graph,
			  NodeNumber  Node,
			  SimpleSet   *AttSet);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns the numbers of all attributes of a node	*/
		/*   which have currently a defined value. The result of*/
		/*   this procedure is unpredictable if there are some	*/
		/*   derived attributes whose value is not up to date.	*/
		/*   (the evaluation of these attributes will not be    */
		/*   triggered automatically).				*/
		/*   For index attributes, AGGetAllDefinedAttributes	*/
		/*   returns all those, which are declared in the scheme*/
		/*   and whose value is currently defined. You can use	*/
		/*   AGGetAllIndexAttributes to get the attributes	*/
		/*   which have not been defined.			*/
		/*							*/
		/*   AttSet must have been created before.		*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   NOT FOUND : the node "Node" does not exist		*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGGetAllNodesWithUndefinedIndex(GraphNumber     Graph,
				AttributeNumber AttNo,
				SimpleSet       *SetOfNodes);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Returns all nodes which have currently invalid     */
		/*   index/key attribute "AttNo". This procedure is for */
		/*   applications which do not use the automatic evalu- */
		/*   facility of GRAS but have their own method to re-  */
		/*   compute invalid derived attributes. These appli-   */
		/*   cations need knowledge about invalid index attri-  */
		/*   butes in order to be able to compute all these     */
		/*   attributes before starting a corresponding index   */
		/*   query.						*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   IS CLOSED : graph "Graph" not opened		*/
		/*   NOT FOUND : an attribute with a number "AttNo" 	*/
		/*		 does not exist.			*/
		/*   UNKNOWN   : the set has not been created		*/
		/*							*/
		/********************************************************/

extern void
AGChangeNodeType(GraphNumber Graph,
		 NodeNumber  Node,
		 TypeNumber  NewType);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Replaces the type of the given node with "NewType".*/
		/*   Attributes, names and edges of the	node are not 	*/
		/*   affected.						*/
		/*							*/
                /*   DON'T USE this procedure within new applications!  */
		/*							*/
		/* Note:						*/
		/*							*/
		/*   This operation is only admissible if the new and	*/
		/*   the old node type have the same attributes and	*/
		/*   evaluation functions (this property won't be	*/
		/*   assured or even checked)! Furthermore, instances   */
		/*   of the new type must be allowed to have the same	*/
		/*   edge context as instances of the old type.		*/
		/*   Last but not least this function does not work for */
		/*   node types	which are declared within the graph	*/
		/*   scheme with "UseMembership"=YES (for efficiency	*/
		/*   reasons only).					*/
		/*							*/
		/* Error Messages:					*/
		/*							*/
		/*   NOT FOUND : the node "Node" does NOT exist		*/
		/*   LOCKED    : the graph has been opened FOR read	*/
		/*		 access only				*/
		/*   IS CLOSED : graph "Graph" NOT opened		*/
		/*							*/
		/********************************************************/


extern void
AGFlushCaches(GraphNumber Graph);
		/********************************************************/
		/*							*/
		/* Purpose:						*/
		/*							*/
		/*   Force Write-Back of client-caches to server	*/
		/*							*/
		/* Note:						*/
		/*							*/
		/*   This is a hack for multi-user mode of gras.	*/
		/*   It is uncritical to use, but slows down the appli- */
		/*   cation. Use it, if you want two clients to coorpe-	*/
		/*   outside transactions (this *is* very dangerous).   */
		/*							*/
		/********************************************************/ 

#ifdef __cplusplus
}
#endif

#endif
